// hlavickove subory *******************************************************
#include "../includes.h"
// vlastny .h subor
#include "frustum.h"

// inicializaica frustum - nacitanie prvkov projekcnej matice **************
void CFrustum::init(void)
{
 // nacitanie prvkov projekcnej matice
 glGetFloatv(GL_PROJECTION_MATRIX,proj);
}
 
// vypocet prvkov rovnic rovin frustum *************************************
void CFrustum::update(void)
{
 // nacitanie prvkov modelview matice
 glGetFloatv(GL_MODELVIEW_MATRIX,modv);

 // pomocne premenne pre vypocet prvkov rovnic rovin frustum - kombinacia matic
 clip[ 0]=modv[ 0]*proj[ 0]+modv[ 1]*proj[ 4]+modv[ 2]*proj[ 8]+modv[ 3]*proj[12];
 clip[ 1]=modv[ 0]*proj[ 1]+modv[ 1]*proj[ 5]+modv[ 2]*proj[ 9]+modv[ 3]*proj[13];
 clip[ 2]=modv[ 0]*proj[ 2]+modv[ 1]*proj[ 6]+modv[ 2]*proj[10]+modv[ 3]*proj[14];
 clip[ 3]=modv[ 0]*proj[ 3]+modv[ 1]*proj[ 7]+modv[ 2]*proj[11]+modv[ 3]*proj[15];

 clip[ 4]=modv[ 4]*proj[ 0]+modv[ 5]*proj[ 4]+modv[ 6]*proj[ 8]+modv[ 7]*proj[12];
 clip[ 5]=modv[ 4]*proj[ 1]+modv[ 5]*proj[ 5]+modv[ 6]*proj[ 9]+modv[ 7]*proj[13];
 clip[ 6]=modv[ 4]*proj[ 2]+modv[ 5]*proj[ 6]+modv[ 6]*proj[10]+modv[ 7]*proj[14];
 clip[ 7]=modv[ 4]*proj[ 3]+modv[ 5]*proj[ 7]+modv[ 6]*proj[11]+modv[ 7]*proj[15];

 clip[ 8]=modv[ 8]*proj[ 0]+modv[ 9]*proj[ 4]+modv[10]*proj[ 8]+modv[11]*proj[12];
 clip[ 9]=modv[ 8]*proj[ 1]+modv[ 9]*proj[ 5]+modv[10]*proj[ 9]+modv[11]*proj[13];
 clip[10]=modv[ 8]*proj[ 2]+modv[ 9]*proj[ 6]+modv[10]*proj[10]+modv[11]*proj[14];
 clip[11]=modv[ 8]*proj[ 3]+modv[ 9]*proj[ 7]+modv[10]*proj[11]+modv[11]*proj[15];

 clip[12]=modv[12]*proj[ 0]+modv[13]*proj[ 4]+modv[14]*proj[ 8]+modv[15]*proj[12];
 clip[13]=modv[12]*proj[ 1]+modv[13]*proj[ 5]+modv[14]*proj[ 9]+modv[15]*proj[13];
 clip[14]=modv[12]*proj[ 2]+modv[13]*proj[ 6]+modv[14]*proj[10]+modv[15]*proj[14];
 clip[15]=modv[12]*proj[ 3]+modv[13]*proj[ 7]+modv[14]*proj[11]+modv[15]*proj[15];

 // prvky rovnice 1. roviny frustum
 plane[0][0]=clip[ 3]-clip[ 0]; plane[0][1]=clip[ 7]-clip[ 4];
 plane[0][2]=clip[11]-clip[ 8]; plane[0][3]=clip[15]-clip[12];
 // prvky rovnice 2. roviny frustum
 plane[1][0]=clip[ 3]+clip[ 0]; plane[1][1]=clip[ 7]+clip[ 4];
 plane[1][2]=clip[11]+clip[ 8]; plane[1][3]=clip[15]+clip[12];
 // prvky rovnice 3. roviny frustum
 plane[2][0]=clip[ 3]+clip[ 1]; plane[2][1]=clip[ 7]+clip[ 5];
 plane[2][2]=clip[11]+clip[ 9]; plane[2][3]=clip[15]+clip[13];
 // prvky rovnice 4. roviny frustum
 plane[3][0]=clip[ 3]-clip[ 1]; plane[3][1]=clip[ 7]-clip[ 5];
 plane[3][2]=clip[11]-clip[ 9]; plane[3][3]=clip[15]-clip[13];
 // prvky rovnice 5. roviny frustum
 plane[4][0]=clip[ 3]-clip[ 2]; plane[4][1]=clip[ 7]-clip[ 6];
 plane[4][2]=clip[11]-clip[10]; plane[4][3]=clip[15]-clip[14];
 // prvky rovnice 6. roviny frustum
 plane[5][0]=clip[ 3]+clip[ 2]; plane[5][1]=clip[ 7]+clip[ 6];
 plane[5][2]=clip[11]+clip[10]; plane[5][3]=clip[15]+clip[14];
}

// vykreslenie frustum *****************************************************
void CFrustum::draw(void)
{
 glDisable(GL_TEXTURE_2D);
 glLineWidth(2.0f);
 glColor4f(1.0f,0.0f,0.0f,1.0f);
 glBegin(GL_LINES);
 loopi(6) {
  glVertex3f(0.0f,0.0f,0.0f); glVertex3f(plane[i][0],plane[i][1],plane[i][2]);
 }
 glEnd();
 glLineWidth(1.0f);
 glColor4f(1.0f,1.0f,1.0f,1.0f);
 glEnable(GL_TEXTURE_2D);
}

/*
  pri teste polohy bodu (trojuholnika, gule, kocky,...) voci frustum
  je vyuzivany fakt, ze normaly vsetkych rovin frustum smeruju dovnutra
  frustum priestoru a teda pokial je bod (aspon jeden z vrcholov
  trojuholnika, gule, kocky,...) pred vsetkymi 6-timi rovinami frustum,
  je vnutri frustum
*/

// test polohy bodu voci frustum *******************************************
bool CFrustum::point_in_frustum(vec3 P)
{
 // test so vsetkymi 6-timi rovinami frustum
 loopi(6) {
  if (plane[i][0]*P.x+plane[i][1]*P.y+plane[i][2]*P.z+plane[i][3]<=0.0f) {
   // bod je za rovinou => je von z frustum
   return(false);
  }
 }
 // bod je pred vsetkymi 6-timi rovinami => je vnutri frustum
 return(true);
}

// test polohy trojuholnika voci frustum ***********************************
bool CFrustum::triangle_in_frustum(vec3 V0,vec3 V1,vec3 V2)
{
 // testuju sa vsetky 3 vrcholy trojuholnika, trojuholnik je von z frustum
 // iba v pripade, ak vsetky jeho 3 vrcholy lezia za rovnakou
 // rovinou frustum, v opacnom pripade je trojuholnik vnutri frustum

 // test so vsetkymi 6-timi rovinami frustum
 loopi(6) {
  if (plane[i][0]*V0.x+plane[i][1]*V0.y+plane[i][2]*V0.z+plane[i][3]>0.0f)
   continue; // ak je vrchol pred rovinou, testuje sa dalsia rovina
  if (plane[i][0]*V1.x+plane[i][1]*V1.y+plane[i][2]*V1.z+plane[i][3]>0.0f)
   continue; // ak je vrchol pred rovinou, testuje sa dalsia rovina
  if (plane[i][0]*V2.x+plane[i][1]*V2.y+plane[i][2]*V2.z+plane[i][3]>0.0f)
   continue; // ak je vrchol pred rovinou, testuje sa dalsia rovina
  // vsetky 3 vrcholy su za rovnakou rovinou => trojuholnik je von z frustum
  return(false);
 }
 // trojuholnik je vnutri frustum
 return(true);
}

// test polohy gule voci frustum *******************************************
bool CFrustum::sphere_in_frustum(vec3 P,GLfloat R)
{
 // test so vsetkymi 6-timi rovinami frustum
 loopi(6) {
  if (plane[i][0]*P.x+plane[i][1]*P.y+plane[i][2]*P.z+plane[i][3]<=-R) {
   // gula je za rovinou => je von z frustum
   return(false);
  }
 }
 // gula je pred vsetkymi 6-timi rovinami => je vnutri frustum
 return(true);
}

// test polohy kocky voci frustum ******************************************
bool CFrustum::cube_in_frustum(vec3 P,GLfloat S)
{
 // testuje sa vsetkych 8 vrcholov kocky, kocka je von z frustum
 // iba v pripade, ak vsetkych jej 8 vrcholov lezi za rovnakou
 // rovinou frustum, v opacnom pripade je kocka vnutri frustum

 // test so vsetkymi 6-timi rovinami frustum
 loopi(6) {
  if (plane[i][0]*(P.x-S)+plane[i][1]*(P.y-S)+plane[i][2]*(P.z-S)+plane[i][3]>0.0f)
   continue; // ak je vrchol pred rovinou, testuje sa dalsia rovina
  if (plane[i][0]*(P.x+S)+plane[i][1]*(P.y-S)+plane[i][2]*(P.z-S)+plane[i][3]>0.0f)
   continue; // ak je vrchol pred rovinou, testuje sa dalsia rovina
  if (plane[i][0]*(P.x+S)+plane[i][1]*(P.y+S)+plane[i][2]*(P.z-S)+plane[i][3]>0.0f)
   continue; // ak je vrchol pred rovinou, testuje sa dalsia rovina
  if (plane[i][0]*(P.x-S)+plane[i][1]*(P.y+S)+plane[i][2]*(P.z-S)+plane[i][3]>0.0f)
   continue; // ak je vrchol pred rovinou, testuje sa dalsia rovina
  if (plane[i][0]*(P.x-S)+plane[i][1]*(P.y-S)+plane[i][2]*(P.z+S)+plane[i][3]>0.0f)
   continue; // ak je vrchol pred rovinou, testuje sa dalsia rovina
  if (plane[i][0]*(P.x+S)+plane[i][1]*(P.y-S)+plane[i][2]*(P.z+S)+plane[i][3]>0.0f)
   continue; // ak je vrchol pred rovinou, testuje sa dalsia rovina
  if (plane[i][0]*(P.x+S)+plane[i][1]*(P.y+S)+plane[i][2]*(P.z+S)+plane[i][3]>0.0f)
   continue; // ak je vrchol pred rovinou, testuje sa dalsia rovina
  if (plane[i][0]*(P.x-S)+plane[i][1]*(P.y+S)+plane[i][2]*(P.z+S)+plane[i][3]>0.0f)
   continue; // ak je vrchol pred rovinou, testuje sa dalsia rovina
  // vsetkych 8 vrcholov je za rovnakou rovinou => kocka je von z frustum
  return(false);
 }
 // kocka je vnutri frustum
 return(true);
}
bool CFrustum::cube_in_frustum(GLfloat MINX,GLfloat MAXX,
                               GLfloat MINY,GLfloat MAXY,
                               GLfloat MINZ,GLfloat MAXZ)
{
 loopi(6) {
  if (plane[i][0]*MINX+plane[i][1]*MINY+plane[i][2]*MINZ+plane[i][3]>0.0f)
   continue;
  if (plane[i][0]*MAXX+plane[i][1]*MINY+plane[i][2]*MINZ+plane[i][3]>0.0f)
   continue;
  if (plane[i][0]*MAXX+plane[i][1]*MAXY+plane[i][2]*MINZ+plane[i][3]>0.0f)
   continue;
  if (plane[i][0]*MINX+plane[i][1]*MAXY+plane[i][2]*MINZ+plane[i][3]>0.0f)
   continue;
  if (plane[i][0]*MINX+plane[i][1]*MINY+plane[i][2]*MAXZ+plane[i][3]>0.0f)
   continue;
  if (plane[i][0]*MAXX+plane[i][1]*MINY+plane[i][2]*MAXZ+plane[i][3]>0.0f)
   continue;
  if (plane[i][0]*MAXX+plane[i][1]*MAXY+plane[i][2]*MAXZ+plane[i][3]>0.0f)
   continue;
  if (plane[i][0]*MINX+plane[i][1]*MAXY+plane[i][2]*MAXZ+plane[i][3]>0.0f)
   continue;
  return(false);
 }
 return(true);
}
