// hlavickove subory *******************************************************
#include "../includes.h"
// vlastny .h subor
#include "loadmd3.h"

int stopaction;

vec3 fireP,fireV; // poloha ustia hlavne a dopredu smerujuci vektor pre strelbu

// test dostupnosti suboru *************************************************
int isFile(char *FNAME)
{
 FILE *f;
 f=fopen(FNAME,"rb");

 if (!f) return(1); // subor nemozno otvorit
 else { fclose(f); return(0); }
}

// pre pracu s retazcom
char str_token[1024];

// nacitanie slova z retazca ***********************************************
char *parseStr(char *DATA)
{
 int c=1,len=0;

 str_token[0]=0;

 if (!DATA) return(NULL);

 while (c!=0) {
  // preskocenie vsetkych netlacitelnych znakov - prvych 32 znakov tabulky
  while ((c=*DATA)<=' ') {
   if (c==0) return(NULL); // koniec retazca
   DATA++;
  }
  // preskocenie komentarov
  if ((c=='/') && (DATA[1])=='/') {
   while ((*DATA) && (*DATA!='\n')) DATA++; // cely riadok
  }
  else break;
 }

 // slovo uzavrete v uvodzovkach
 if (c=='\"') {
  DATA++;
  while ((c=*DATA)!=0) {
   if (c=='\"') { str_token[len]=0; return(DATA); }
   str_token[len]=c; len++;
   DATA++;
  }
 }

 // jeden znak
 if ((c=='{') || (c=='}') || (c==')') || (c=='(') ||
     (c=='\'') || (c==':') || (c==',') || (c=='/')) {
  str_token[len]=c; len++; str_token[len]=0;
  return(DATA+1);
 }

 // jedno slovo
 while (c>32) {
  str_token[len]=c; len++;
  DATA++;
  c=*DATA;
  if ((c=='{') || (c=='}') || (c==')') || (c=='(') ||
      (c=='\'') || (c==':') || (c==',') || (c=='/')) break;
 }
 str_token[len]=0;

 return(DATA);
}

// retazec je/nie je cislo *************************************************
int isNum(char *DATA)
{
 while (*DATA) {
  if ((*DATA<'0') || (*DATA>'9')) return(0);
  DATA++;
 }

 return(1);
}

// CLoadMD3 ****************************************************************

// konstruktor *************************************************************
CLoadMD3::CLoadMD3(void)
{
 fp=NULL; // subor

 frames=NULL;   // snimky
 tags=NULL;     // znacky
 surfaces=NULL; // sietove modely
 links=NULL;    // prepojenia

 // dekodovanie normal
 float lat,lng;
 // priprava tabulky s predpripravenymi normalami
 loopj(256) {
  loopi(256) {
   lat=i*2.0f*M_PI/255.0f;
   lng=j*2.0f*M_PI/255.0f;

   nxyz[i][j][0]=(float) (cos(lat)*sin(lng));
   nxyz[i][j][1]=(float) (sin(lat)*sin(lng));
   nxyz[i][j][2]=(float) cos(lng);
  }
 }

 actframe=0; nextframe=0;
 ipol=0.0f;
 stopaction=1; // Avatar nestriela

 actV=0; ofsV=0; ipos=0;
}

// destruktor **************************************************************
CLoadMD3::~CLoadMD3(void)
{
 DELETEA(frames);      // snimky
 DELETEA(tags);        // znacky
 if (surfaces!=NULL) { // sietove modely
  loopk(header.numSurfaces) {
   DELETEA(surfaces[k].shaders); DELETEA(surfaces[k].triangles);
   DELETEA(surfaces[k].texcoords); DELETEA(surfaces[k].vertices);

   DELETEA(surfaces[k].tris_v); // pole trojuholnikov
   DELETEA(surfaces[k].tris_t); // pole texturovacich suradnic
  }
 }
 DELETEA(surfaces);
 DELETEA(links);       // prepojenia
}

// nacitanie obsahu MD3 suboru *********************************************
int CLoadMD3::load(const char *FNAME)
{
 // otvorenie MD3 suboru
 fp=fopen(FNAME,"rb");
 if (fp==NULL) { printf("Nemozno otvorit MD3 subor: %s!\n",FNAME); return(1); }

 long cur_offset=ftell(fp); // aktualna pozicia v subore

 // nacitanie hlavicky MD3 suboru
 fr=fread(&header,sizeof(MD3header),1,fp);
 // test verzie MD3 suboru
 if (header.version!=15) {
  printf("Nepodporovana verzia MD3 suboru: %i!\n",header.version); fclose(fp); return(1);
 }
 // INFO: MD3 subor
 printf("MD3 subor: %s\n",FNAME);
 // INFO: hlavicka MD3 suboru
 printf("ident: %c%c%c%c ",header.ident[0],header.ident[1],header.ident[2],header.ident[3]);
 printf("name: %s ",header.name);
 printf("frames: %i ",header.numFrames);
 printf("tags: %i ",header.numTags);
 printf("surfaces: %i ",header.numSurfaces);
 printf("skins: %i\n",header.numSkins);

 // vyhradenie pamate
 frames=(MD3frame *) malloc(sizeof(MD3frame)*header.numFrames);
 tags=(MD3tag *) malloc(sizeof(MD3tag)*header.numFrames*header.numTags);
 surfaces=(MD3surface *) malloc(sizeof(MD3surface)*header.numSurfaces);
 links=(CLoadMD3 **) malloc(sizeof(CLoadMD3 *)*header.numTags);
 if ((frames==NULL)||(tags==NULL)||(surfaces==NULL)||(links==NULL)) {
  printf("Nemozno vyhradit potrebnu pamat!\n"); fclose(fp); return(1);
 }

 // nacitanie snimok
 fseek(fp,cur_offset+header.ofsFrames,SEEK_SET);
 fr=fread(frames,sizeof(MD3frame),header.numFrames,fp);

 a=DEG2RAD(90.0f); // a = 90 stupnov v radianoch

 // nacitanie znaciek
 fseek(fp,cur_offset+header.ofsTags,SEEK_SET);
 fr=fread(tags,sizeof(MD3tag),header.numFrames*header.numTags,fp);
 // INFO: znacky MD3 suboru
 if (header.numTags!=0) {
  printf("znacky: "); loopi(header.numTags) printf("%i %s ",i,tags[i].name); printf("\n");
  // pre vsetky snimky vsetkych znaciek
  loopi(header.numFrames*header.numTags) {
   tags[i].origin[0]/=2000.0f/64.0f;
   tags[i].origin[1]/=2000.0f/64.0f;
   tags[i].origin[2]/=2000.0f/64.0f;
   tags[i].origin[0]*=1.3f; tags[i].origin[1]*=1.3f; tags[i].origin[2]*=1.3f; // skalovanie
   // rotacia lokalnej polohy
   u1.set(tags[i].origin[0],tags[i].origin[1],tags[i].origin[2]);
   pV.set(0.0f,0.0f,1.0f); u1.rotate(a,pV);
   pV.set(-1.0f,0.0f,0.0f); u1.rotate(a,pV);
   tags[i].origin[0]=u1.x;
   tags[i].origin[1]=u1.y;
   tags[i].origin[2]=u1.z;
   // rotacia lokalnych natoceni
   u1.set(tags[i].axis[0][0],tags[i].axis[0][1],tags[i].axis[0][2]);
   u2.set(tags[i].axis[1][0],tags[i].axis[1][1],tags[i].axis[1][2]);
   u3.set(tags[i].axis[2][0],tags[i].axis[2][1],tags[i].axis[2][2]);
   pV.set(0.0f,0.0f,1.0f);
   u1.rotate(a,pV); u2.rotate(a,pV); u3.rotate(a,pV);
   pV.set(-1.0f,0.0f,0.0f);
   u1.rotate(a,pV); u2.rotate(a,pV); u3.rotate(a,pV);
   tags[i].axis[0][0]=u1.x; tags[i].axis[0][1]=u1.y; tags[i].axis[0][2]=u1.z;
   tags[i].axis[1][0]=u2.x; tags[i].axis[1][1]=u2.y; tags[i].axis[1][2]=u2.z;
   tags[i].axis[2][0]=u3.x; tags[i].axis[2][1]=u3.y; tags[i].axis[2][2]=u3.z;
  } // pre vsetky snimky vsetkych znaciek
 } // pre vsetky znacky

 // nacitanie sietovych modelov
 cur_offset+=header.ofsSurfaces;
 loopi(header.numSurfaces) { // pre vsetky sietove modely
  fseek(fp,cur_offset,SEEK_SET);
  // hlavicka sietoveho modelu
  fr=fread(&surfaces[i].header,sizeof(MD3surface_header),1,fp);
  // vyhradenie pamate
  surfaces[i].shaders=(MD3shader *) malloc(sizeof(MD3shader)*surfaces[i].header.numShaders);
  surfaces[i].triangles=(MD3triangle *) malloc(sizeof(MD3triangle)*surfaces[i].header.numTriangles);
  surfaces[i].texcoords=(MD3st *) malloc(sizeof(MD3st)*surfaces[i].header.numVerts);
  surfaces[i].vertices=(MD3vertex *) malloc(sizeof(MD3vertex)*surfaces[i].header.numFrames*surfaces[i].header.numVerts);
  if ((surfaces[i].shaders==NULL)||(surfaces[i].triangles==NULL)||
      (surfaces[i].texcoords==NULL)||(surfaces[i].vertices==NULL)) {
   printf("Nemozno vyhradit potrebnu pamat!\n"); fclose(fp); return(1);
  }
  // shadery
  fseek(fp,cur_offset+surfaces[i].header.ofsShaders,SEEK_SET);
  fr=fread(surfaces[i].shaders,sizeof(MD3shader),surfaces[i].header.numShaders,fp);
  // odkazy na vrcholy trojuholnikov
  fseek(fp,cur_offset+surfaces[i].header.ofsTriangles,SEEK_SET);
  fr=fread(surfaces[i].triangles,sizeof(MD3triangle),surfaces[i].header.numTriangles,fp);
  // texturovacie suradnice
  fseek(fp,cur_offset+surfaces[i].header.ofsSt,SEEK_SET);
  fr=fread(surfaces[i].texcoords,sizeof(MD3st),surfaces[i].header.numVerts,fp);
  // vrcholy
  fseek(fp,cur_offset+surfaces[i].header.ofsXyznormal,SEEK_SET);
  fr=fread(surfaces[i].vertices,sizeof(MD3vertex),surfaces[i].header.numFrames*surfaces[i].header.numVerts,fp);
  // presun na dalsi sietovy model
  cur_offset+=surfaces[i].header.ofsEnd;

  // presun trojuholnikov a texturovacich suradnic do poli
  surfaces[i].tris_v=(vec3 *) malloc(3*sizeof(vec3)*surfaces[i].header.numFrames*surfaces[i].header.numTriangles);
  surfaces[i].tris_t=(vec2 *) malloc(3*sizeof(vec2)*surfaces[i].header.numTriangles);
  if ((surfaces[i].tris_v==NULL)||(surfaces[i].tris_t==NULL)) {
   printf("Nemozno vyhradit potrebnu pamat!\n"); return(1);
  }
  printf("surface: %i, triangles: %i\n",i,surfaces[i].header.numTriangles);

  ipos=0; // pozicia v poli trojuholnikov
  loopk(surfaces[i].header.numFrames) { // pre vsetky snimky
   // 1. vrchol zo sady vrcholov konkretneho sietoveho modelu patriacich snimke k
   ofsV=k*surfaces[i].header.numVerts;

   loopj(surfaces[i].header.numTriangles) { // pre vsetky trojuholniky
    actV=ofsV+surfaces[i].triangles[j].indexes[0]; // odkaz na 1. vrchol trojuholnika
    surfaces[i].tris_v[ipos+0].set((float) (surfaces[i].vertices[actV].v[0])/2000.0f,
                                   (float) (surfaces[i].vertices[actV].v[1])/2000.0f,
                                   (float) (surfaces[i].vertices[actV].v[2])/2000.0f);
    actV=ofsV+surfaces[i].triangles[j].indexes[1]; // odkaz na 2. vrchol trojuholnika
    surfaces[i].tris_v[ipos+2].set((float) (surfaces[i].vertices[actV].v[0])/2000.0f,
                                   (float) (surfaces[i].vertices[actV].v[1])/2000.0f,
                                   (float) (surfaces[i].vertices[actV].v[2])/2000.0f);
    actV=ofsV+surfaces[i].triangles[j].indexes[2]; // odkaz na 3. vrchol trojuholnika
    surfaces[i].tris_v[ipos+1].set((float) (surfaces[i].vertices[actV].v[0])/2000.0f,
                                   (float) (surfaces[i].vertices[actV].v[1])/2000.0f,
                                   (float) (surfaces[i].vertices[actV].v[2])/2000.0f);
    // rotacia vrcholov trojuholnikov
    pV.set(0.0f,0.0f,1.0f);
    surfaces[i].tris_v[ipos+0].rotate(a,pV);
    surfaces[i].tris_v[ipos+1].rotate(a,pV);
    surfaces[i].tris_v[ipos+2].rotate(a,pV);
    pV.set(-1.0f,0.0f,0.0f);
    surfaces[i].tris_v[ipos+0].rotate(a,pV);
    surfaces[i].tris_v[ipos+1].rotate(a,pV);
    surfaces[i].tris_v[ipos+2].rotate(a,pV);
    // skalovanie
    surfaces[i].tris_v[ipos+0]*=1.3f;
    surfaces[i].tris_v[ipos+1]*=1.3f;
    surfaces[i].tris_v[ipos+2]*=1.3f;
    ipos+=3; // posun v poli trojuholnikov
   } // pre vsetky trojuholniky
  } // pre vsetky snimky

  ipos=0; // pozicia v poli texturovacich suradnic
  loopj(surfaces[i].header.numTriangles) { // pre vsetky trojuholniky
   actV=surfaces[i].triangles[j].indexes[0]; // odkaz na 1. vrchol trojuholnika
   surfaces[i].tris_t[ipos+0].set(surfaces[i].texcoords[actV].st[0],surfaces[i].texcoords[actV].st[1]);
   actV=surfaces[i].triangles[j].indexes[1]; // odkaz na 2. vrchol trojuholnika
   surfaces[i].tris_t[ipos+2].set(surfaces[i].texcoords[actV].st[0],surfaces[i].texcoords[actV].st[1]);
   actV=surfaces[i].triangles[j].indexes[2]; // odkaz na 3. vrchol trojuholnika
   surfaces[i].tris_t[ipos+1].set(surfaces[i].texcoords[actV].st[0],surfaces[i].texcoords[actV].st[1]);
   ipos+=3; // posun v poli texturovacich suradnic
  } // pre vsetky trojuholniky

  // normaly
  //au=(int) surfaces[k].vertices[ofsV+actV].n[0];
  //av=(int) surfaces[k].vertices[ofsV+actV].n[1];
  // dekodovanie normal
  //n[0]=nxyz[au][av][0];
  //n[1]=nxyz[au][av][1];
  //n[2]=nxyz[au][av][2];

  // uvolnenie vyhradenej pamate - tieto udaje uz dalej nebudu potrebne
  DELETEA(surfaces[i].shaders); DELETEA(surfaces[i].triangles);
  DELETEA(surfaces[i].texcoords); DELETEA(surfaces[i].vertices);
 } // pre vsetky sietove modely

 // reset prepojeni
 loopi(header.numTags) links[i]=NULL;

 // posledna snimka pre Q3 model je header.numFrames-1
 header.numFrames--;

 // zatvorenie MD3 suboru
 fclose(fp);

 // nahratie textury - iba pre 1. sietovy model
 char tmpName[255]; // pomocna premenna
 strcpy(tmpName,FNAME);
 ipos=strlen(tmpName)-1; while (tmpName[ipos]!='/') ipos--;
 // textura zbrane
 tmpName[ipos]='\0'; strcat(tmpName,"/skin.jpg\0");
 texture.load2D(tid[3],tmpName,T_MIPMAP|T_LINEAR);
 // textura dier po gulkach
 tmpName[ipos]='\0'; strcat(tmpName,"/holes_concrete.png\0");
 texture.load2D(tid[4],tmpName,T_MIPMAP|T_LINEAR|T_CLAMP_TO_EDGE);

 return(0);
}

// nacitanie obrazkov pre povrchy sietovych modelov ************************
int CLoadMD3::loadSkin(char *FNAME,char *IPATH)
{
 FILE *f;
 char *text_line,*buf;
 char name_surface[1024],name_image[1024],path[1024];
 unsigned char issurface,istag,isimage;

 if (isFile(FNAME)) {
  fprintf(stderr,"Nemozno najst subor %s!\n",FNAME); return(1);
 }
 // otvorenie suboru
 f=fopen(FNAME,"rt");
 buf=(char *) malloc(1024);
 // nacitanie obrazkov
 while (!feof(f)) {
  istag=0;
  issurface=0; isimage=0;
  text_line=buf;
  text_line[0]='\0';
  int ret=fscanf(f,"%s",text_line);
  if (ret!=0) ;

  if (text_line && text_line[0]) {
   // nacitanie znacky,nazvu povrchu,nazvu obrazka
   while (text_line) {
    text_line=parseStr(text_line);

    if (strstr(str_token,"tag_") && issurface==0 && istag==0) { istag=1; }
    else if (issurface==0 && istag==0) { strcpy(name_surface,str_token); issurface=1; }
         else if ((strstr(str_token,".tga")||strstr(str_token,".jpg")) && issurface==1 && istag==0)
              { strcpy(name_image,str_token); isimage=1; }
   }

   // nacitanie obrazka
   if (issurface && isimage) {
    fprintf(stderr,"%s...%s\n",name_surface,name_image);

    if (strstr(name_image,".tga") || strstr(name_image,".jpg")) {
     sprintf(path,"data/models/%s/%s",IPATH,name_image);

     i=texture.load2D(tid[texture.tex_count-1],path,T_LINEAR|T_MIPMAP);
     // nastavenie textury pre povrch sietoveho modelu
     if (!i) setSurfaceTexture(name_surface,texture.tex_id[texture.tex_count-2]);
    }
   }
  }
 }

 free(buf);
 // zatvorenie suboru
 fclose(f);

 return(0);
}

// nastavenie textury pre povrchy sietovych modelov ************************
int CLoadMD3::setSurfaceTexture(char *S,unsigned int TID)
{
 loopi(header.numSurfaces) {
  if (!strcmp((char *) surfaces[i].header.name,S)) {
   //fprintf(stderr,"surface: %s texture id: %u\n",surfaces[i].header.name,TID);
   surfaces[i].tex_id=TID;
   surfaces[i].is_texture=1;
   return(i);
  }
 }
 return(0);
}

// prepojenia modelov ******************************************************
int CLoadMD3::linkModel(char *T,CLoadMD3 *M)
{
 // v pripade,ze sa nazov znacky rovna zadanemu nazvu,
 // k prepojeniam aktualneho modelu sa doplni odkaz na novy model

 // pre vsetky znacky aktualneho modelu
 loopi(header.numTags) {
  // ak sa nazov znacky rovna zadanemu nazvu
  if (!strcmp((char *) tags[i].name,T)) {
   // vznikne nove prepojenie
   links[i]=M;
   return(i);
  }
 }

 return(0);
}

// nastavenie informacii o snimku ******************************************
void CLoadMD3::setFrameInfo(int F,int N,float P)
{
 actframe=F;
 nextframe=N;

 if (P<0.0f || P>1.0f) P=0.0f;
 ipol=P;
}

// ziskanie informacii o snimku ********************************************
void CLoadMD3::getFrameInfo(int *F,int *N,float *P)
{
 *F=actframe;
 *N=nextframe;
 *P=ipol;
}

// vykreslenie 1 konkretnej snimky *****************************************
void CLoadMD3::drawModel1F(int AF)
{
 glMatrixMode(GL_MODELVIEW); // modelview matica
 glPushMatrix();             // ulozenie
 glLoadIdentity();           // reset

  // textura zbrane
 glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[3]]);

 glEnableClientState(GL_VERTEX_ARRAY);
 glEnableClientState(GL_TEXTURE_COORD_ARRAY);

 loopk(header.numSurfaces) { // pre vsetky sietove modely
  glVertexPointer(3,GL_FLOAT,0,(surfaces[k].tris_v+3*AF*surfaces[k].header.numTriangles));
  glTexCoordPointer(2,GL_FLOAT,0,surfaces[k].tris_t);

  glDrawArrays(GL_TRIANGLES,0,3*surfaces[k].header.numTriangles);
 } // pre vsetky sietove modely

 glDisableClientState(GL_VERTEX_ARRAY);
 glDisableClientState(GL_TEXTURE_COORD_ARRAY);

 glPopMatrix();              // modelview matica
}

vec3 EV,TV;    // poloha ustia hlavne a dopredu smerujuci vektor
vec3 AA,BB,CC; // vrcholy zasiahnuteho trojuholnika - iba pre INFO
bool impact;   // indikator zasahu protihraca

// vykreslenie interpolovanej snimky ***************************************
void CLoadMD3::drawModelIF(int AF)
{
 glEnableClientState(GL_VERTEX_ARRAY);
 glEnableClientState(GL_TEXTURE_COORD_ARRAY);

 loopk(header.numSurfaces) { // pre vsetky sietove modely
  if (surfaces[k].is_texture) glBindTexture(GL_TEXTURE_2D,surfaces[k].tex_id);

  glVertexPointer(3,GL_FLOAT,0,(surfaces[k].tris_v+3*AF*surfaces[k].header.numTriangles));
  glTexCoordPointer(2,GL_FLOAT,0,surfaces[k].tris_t);

  glDrawArrays(GL_TRIANGLES,0,3*surfaces[k].header.numTriangles);
 } // pre vsetky sietove modely

 glDisableClientState(GL_VERTEX_ARRAY);
 glDisableClientState(GL_TEXTURE_COORD_ARRAY);

 // ak sa nema testovat zasah trojuholnika gulkou, funkcia sa ukonci
 if (TV.x==2.0f) return;

 // transformacna matica, pomocou ktorej sa transformuje aktualne vykreslovany objekt
 // z OBJECT do WORLD space
 glGetFloatv(GL_MODELVIEW_MATRIX,transfMAT.get());

 minr=wRange; // maximalny dostrel

 loopk(header.numSurfaces) { // pre vsetky sietove modely

  // ukazovatel na 1. trojuholnik konkretneho sietoveho modelu konkretnej snimky
  triV=surfaces[k].tris_v+3*AF*surfaces[k].header.numTriangles;

  i=0;
  loopl(surfaces[k].header.numTriangles) { // pre vsetky trojuholniky
   u1=transfMAT*triV[i+0]; // transformacia 1. vrcholu
   u2=transfMAT*triV[i+1]; // transformacia 2. vrcholu
   u3=transfMAT*triV[i+2]; // transformacia 3. vrcholu
   i+=3;

   // nasledujuce vypocty normal je vhodnejsie vykonat vopred pri nacitani MD3 postavy
   // normaly mozu byt ulozene spolu s trojuholnikmi pre kazdu snimku
   // sice by to znamenalo zvacsenie udajoveho priestoru avsak zvysil by sa tym vykon aplikacie
   // vypocet normaly trojuholnika
   a0=u2-u1; a1=u3-u1;
   // vektorovy sucin dvoch vektorov = normala trojuholnika
   a2=a0.cross(a1); a2.normalize();
   trisnormal=a2;
   // vypocet prvku D rovnice roviny, ktoru tvori trojuholnik
   // D=-(AX+BY+CZ)
   // A,B,C - normala trojuholnika
   // X,Y,Z - lubovolny vrchol trojuholnika
   trisd=-(a2.x*u1.x+a2.y*u1.y+a2.z*u1.z);

   // vypocet koliznych normal
   a1=trisnormal; // normala trojuholnika
   loopj(3) {
    if (j==0) a0=u2-u1; // 1. strana trojuholnika
    if (j==1) a0=u3-u2; // 2. strana trojuholnika
    if (j==2) a0=u1-u3; // 3. strana trojuholnika
    // vektorovy sucin vektorov = kolizna normala
    // prvky koliznej normaly su prvkami rovnice roviny
    // A,B,C, nasledne sa vypocita prvok D
    a2=a0.cross(a1); a2.normalize();
    triscn[j].x=a2.x; triscn[j].y=a2.y; triscn[j].z=a2.z;
    // z rovnice roviny AX+BY+CZ+D=0 => D=-(AX+BY+CZ)
    if (j==0) triscn[j].w=-(a2.x*u1.x+a2.y*u1.y+a2.z*u1.z);
    if (j==1) triscn[j].w=-(a2.x*u2.x+a2.y*u2.y+a2.z*u2.z);
    if (j==2) triscn[j].w=-(a2.x*u3.x+a2.y*u3.y+a2.z*u3.z);
   }

   // ak luc zasiahol trojuholnik
   if (m3d.ray_tri_collision(-EV,-TV,trisnormal,trisd,triscn[0],triscn[1],triscn[2])) {
    pV=m3d.get_ipoint(); // priesecnik luca a trojuholnika
/*
    // INFO: priesecnik luca a trojuholnika
    glPushMatrix(); glLoadIdentity();
    glDisable(GL_TEXTURE_2D);
    glPointSize(5.0f);
    glColor4f(1.0f,0.0f,0.0f,1.0f);
    glBegin(GL_POINTS);
     glVertex3fv(pV.get_data());
    glEnd();
    glPointSize(1.0f);
    glEnable(GL_TEXTURE_2D);
    glPopMatrix();
*/
    // vzdialenost priesecnika od ustia hlavne
    maxr=(pV+EV).magnitude();
    if (maxr<minr) { // vyber najblizsieho priesecnika
     ipoint=pV; inormal=trisnormal; minr=maxr;
     AA=u1; BB=u2; CC=u3; // vrcholy zasiahnuteho trojuholnika
     impact=true; // protihrac bol zasiahnuty
    }
   }

  } // pre vsetky trojuholniky
 } // pre vsetky sietove modely

/*
 // pre vsetky povrchy
 for (cur_surface=0;cur_surface<header.numSurfaces;cur_surface++) {
  // posun na prvy vrchol
  cur_offset_vertex=curFrame*surfaces[cur_surface].surface_h.numVerts;
  // interpolacia
  next_cur_offset_vertex=nextFrame*surfaces[cur_surface].surface_h.numVerts;
  // pocet trojuholnikov
  tri_num=surfaces[cur_surface].surface_h.numTriangles;
  // povrch ma texturu
  if (surfaces[cur_surface].is_texture) glBindTexture(GL_TEXTURE_2D,surfaces[cur_surface].tex_id);
  //fprintf(stderr,"surface: %s texture id: %u\n",surfaces[cur_surface].surface_h.name,surfaces[cur_surface].tex_id);

  // pre vsetky trojuholniky
  for (i=0;i<tri_num;i++) {
   glBegin(GL_TRIANGLES);
   // pre vsetky vrcholy trojuholnikov
   for (j=0;j<3;j++) {
    // index vrcholu
    cur_vertex=surfaces[cur_surface].triangles[i].indexes[j];
    // texturovacie suradnice
    s=surfaces[cur_surface].texcoords[cur_vertex].st[0];
    t=surfaces[cur_surface].texcoords[cur_vertex].st[1];
    // normaly
    au=(int) surfaces[cur_surface].vertices[cur_offset_vertex+cur_vertex].n[0];
    av=(int) surfaces[cur_surface].vertices[cur_offset_vertex+cur_vertex].n[1];
    n[0]=nxyz[au][av][0];
    n[1]=nxyz[au][av][1];
    n[2]=nxyz[au][av][2];
     // interpolacia
     au=(int) surfaces[cur_surface].vertices[next_cur_offset_vertex+cur_vertex].n[0];
     av=(int) surfaces[cur_surface].vertices[next_cur_offset_vertex+cur_vertex].n[1];
     next_n[0]=nxyz[au][av][0];
     next_n[1]=nxyz[au][av][1];
     next_n[2]=nxyz[au][av][2];
    //  vrcholy
    v[0]=(float) surfaces[cur_surface].vertices[cur_offset_vertex+cur_vertex].v[0]/64.0f;
    v[1]=(float) surfaces[cur_surface].vertices[cur_offset_vertex+cur_vertex].v[1]/64.0f;
    v[2]=(float) surfaces[cur_surface].vertices[cur_offset_vertex+cur_vertex].v[2]/64.0f;
     // interpolacia
     next_v[0]=(float) surfaces[cur_surface].vertices[next_cur_offset_vertex+cur_vertex].v[0]/64.0f;
     next_v[1]=(float) surfaces[cur_surface].vertices[next_cur_offset_vertex+cur_vertex].v[1]/64.0f;
     next_v[2]=(float) surfaces[cur_surface].vertices[next_cur_offset_vertex+cur_vertex].v[2]/64.0f;

    glTexCoord2f(s,t);
    glNormal3f(n[0]+p*(next_n[0]-n[0]),
               n[1]+p*(next_n[1]-n[1]),
               n[2]+p*(next_n[2]-n[2]));
    glVertex3f(v[0]+p*(next_v[0]-v[0]),
               v[1]+p*(next_v[1]-v[1]),
               v[2]+p*(next_v[2]-v[2]));
   }

   glEnd();
  }
 }
*/
}

// vykreslenie MD3 modelu **************************************************
void CLoadMD3::draw(void)
{
 drawModelIF(actframe); // vykreslenie interpolovanej snimky
}

void CLoadMD3::draw(vec3 EP,vec3 TP,vec3 RP,GLubyte F)
{/*
 if ((F!=0)&&(stopaction==0)) { // ak je zbran v pokoji, avatar moze vystrelit
  stopaction=1; // zbran uz nie je v pokoji => avatar nemoze znovu vystrelit

  // pridanie odtlacku (decal) --------------------------------------------

  // normalizovane vektory smerujuce dopredu, doprava a hore
  frontV=TP; rightV=RP; upV=frontV.cross(rightV);

  ip0=wRange;      // maximalny dostrel
  TP=EP+TP*wRange; // bod na pozicii maximalneho dostrelu

  // pre vsetky trojuholniky
  loopvk(tris) {
   // ak luc zasiahol trojuholnik
   if (m3d.ray_tri_collision(-EP,-TP,tris[k].n,tris[k].d,tris[k].cn[0],tris[k].cn[1],tris[k].cn[2])) {
    pV=m3d.get_ipoint(); // priesecnik luca a trojuholnika
    ip1=(pV+EP).magnitude();
    if (ip1<ip0) { // vyber najblizsieho priesecnika
     ipoint=pV; inormal=tris[k].n; ip0=ip1;
    }
   }
  }

  // ak avatar zasiahol nejaky trojuholnik
  if (ip0!=wRange) {
   // priprava 6-tich orezavacich rovin (steny 3D hranola)
   cplane[0].x=rightV.x; cplane[0].y=rightV.y; cplane[0].z=rightV.z;    // prava stena
   pV=ipoint+rightV*wDecalHalfSize;
   cplane[0].w=-(cplane[0].x*pV.x+cplane[0].y*pV.y+cplane[0].z*pV.z);

   cplane[1].x=-rightV.x; cplane[1].y=-rightV.y; cplane[1].z=-rightV.z; // lava stena
   pV=ipoint-rightV*wDecalHalfSize;
   cplane[1].w=-(cplane[1].x*pV.x+cplane[1].y*pV.y+cplane[1].z*pV.z);

   cplane[2].x=upV.x; cplane[2].y=upV.y; cplane[2].z=upV.z;             // vrchna stena
   pV=ipoint+upV*wDecalHalfSize;
   cplane[2].w=-(cplane[2].x*pV.x+cplane[2].y*pV.y+cplane[2].z*pV.z);

   cplane[3].x=-upV.x; cplane[3].y=-upV.y; cplane[3].z=-upV.z;          // spodna stena
   pV=ipoint-upV*wDecalHalfSize;
   cplane[3].w=-(cplane[3].x*pV.x+cplane[3].y*pV.y+cplane[3].z*pV.z);

   cplane[4].x=frontV.x; cplane[4].y=frontV.y; cplane[4].z=frontV.z;    // zadna stena
   pV=ipoint+frontV*wDecalHalfSize;
   cplane[4].w=-(cplane[4].x*pV.x+cplane[4].y*pV.y+cplane[4].z*pV.z);

   cplane[5].x=-frontV.x; cplane[5].y=-frontV.y; cplane[5].z=-frontV.z; // predna stena
   pV=ipoint-frontV*wDecalHalfSize;
   cplane[5].w=-(cplane[5].x*pV.x+cplane[5].y*pV.y+cplane[5].z*pV.z);

   addDecal(cplane); // pridanie odtlacku (vsetkych casti odtlacku)
  } // ak avatar zasiahol nejaky trojuholnik
 }

 actframe+=max(1,(int) ceil(50.0f*difTime))*stopaction;
 if (actframe>header.numFrames) { actframe=0; stopaction=0; }
 //drawModel1F(actframe); // vykreslenie 1 konkretnej snimky
 */
}

// CPlayer *****************************************************************

// konstruktor *************************************************************
CPlayer::CPlayer()
{
 link=NULL;

 // animacie
 loopi(MAX_ANIMATIONS) {
  anim[i].firstF=0; anim[i].numF=0; anim[i].loopingF=0;
  anim[i].fps=0;
 }

 act_l_anim=0;
 act_u_anim=0;

 // plynutie casu
 aTime=0.0f;
 l_curFrame=0; u_curFrame=0;
 l_nextFrame=0; u_nextFrame=0;
 l_curTime=0.0f; u_curTime=0.0f;
 l_oldTime=0.0f; u_oldTime=0.0f;

 pos=NULL; rot=NULL; next_pos=NULL; next_rot=NULL;

 // prvy, posledny, aktualny odtlacok (casti odtlacku),
 // oznacenie casti posledneho pridaneho odtlacku
 firstD=NULL; lastD=NULL; curD=NULL; cD=NULL; lastG=NULL;
}

// destruktor **************************************************************
CPlayer::~CPlayer()
{
 // zmazanie vsetkych odtlackov (vsetkych casti)
 delAllDecals();
}

// inicializacia postavy ***************************************************
void CPlayer::init(int MODEL)
{
 char player[7]="sarge",skin[]="default",weapon[8]="railgun";

 if (MODEL==1) { strcpy(player,"reaper"); strcpy(weapon,"xael"); }

 // nacitanie postavy + textur + zbrane
 loadPlayer(player,skin,weapon);
}

// nacitanie postavy + textur + zbrane *************************************
void CPlayer::loadPlayer(char *D,char *S,char *W)
{
 char l_mpath[1024],u_mpath[1024],h_mpath[1024]; // model
 char l_spath[1024],u_spath[1024],h_spath[1024]; // skin
 char w_mpath[1024],w_spath[1024];               // weapon
 char a_path[1024];                              // animation

 // casti postavy - MD3 modely (lower,upper,head,weapon) - model
 sprintf(l_mpath,"data/models/%s/lower.md3",D);
 sprintf(u_mpath,"data/models/%s/upper.md3",D);
 sprintf(h_mpath,"data/models/%s/head.md3",D);
 // nacitanie modelov
 l.load(l_mpath); u.load(u_mpath); h.load(h_mpath);

 // povrchy jednotlivych casti - skin
 sprintf(l_spath,"data/models/%s/lower_%s.skin",D,S);
 sprintf(u_spath,"data/models/%s/upper_%s.skin",D,S);
 sprintf(h_spath,"data/models/%s/head_%s.skin",D,S);
 // nacitanie povrchov
 l.loadSkin(l_spath,D); u.loadSkin(u_spath,D); h.loadSkin(h_spath,D);

 // zbran - weapon
 if (W!=NULL) {
  sprintf(w_mpath,"data/models/%s/%s.md3",W,W);
  sprintf(w_spath,"data/models/%s/%s.skin",W,W);
  w.load(w_mpath); w.loadSkin(w_spath,W);
 }

 char torso[]="tag_torso",head[]="tag_head",weapon[]="tag_weapon";
 // prepojenia modelov - links
 l.linkModel(torso,&u);               // lower -> upper (naviazany na tag_torso)
 u.linkModel(head,&h);                // upper -> head (naviazany na tag_head)
 if (W!=NULL) u.linkModel(weapon,&w); // upper -> weapon (naviazany na tag_weapon)
/*
 // INFO: prepojenia medzi modelmi
 printf("lower:\n");
 loopi(l.header.numTags) {
  if (l.links[i]) printf("link %u: P",i); else printf("link %u: NULL",i); printf("\n");
 }
 printf("upper:\n");
 loopi(u.header.numTags) {
  if (u.links[i]) printf("link %u: P",i); else printf("link %u: NULL",i); printf("\n");
 }
 printf("head:\n");
 loopi(h.header.numTags) {
  if (h.links[i]) printf("link %u: P",i); else printf("link %u: NULL",i); printf("\n");
 }
*/
 // animacia - animation
 sprintf(a_path,"data/models/%s/animation.cfg",D);
 // nacitanie animacie
 loadAnim(a_path);
 // nastavenie parametrov animacie
 setAnim(LEGS_IDLE,TORSO_STAND);
}

// nacitanie animacie ******************************************************
int CPlayer::loadAnim(char *F)
{
 FILE *f;
 char *text_line,*buffer;
 int c;
 int a=0,t=0;
 int issex,isheadoffset,isfootsteps,skip;

 if (isFile(F)) {
  fprintf(stderr,"Nemozno najst subor %s!\n",F); return(1);
 }
 // otvorenie suboru
 f=fopen(F,"rt");
 buffer=(char *) malloc(1024);
 // nacitanie obsahu suboru
 while (!feof(f)) {
  c=fgetc(f);

  i=0;
  text_line=buffer;
  text_line[0]=0;
  // preskocenie 10=new line,32=space,-1=EOF
  while (c!=10 && c!=32 && c!=-1) {
   text_line[i]=c; i++;
   c=fgetc(f);
  }
  text_line[i]='\0';

  t=0; // 4 ciselne udaje patriace animacii
  issex=0;
  isheadoffset=0;
  isfootsteps=0;
  while (text_line) {
   text_line=parseStr(text_line);
   if (str_token[0]!=0) {
    if (strstr(str_token,"sex")) issex=1;
    if (strstr(str_token,"headoffset")) isheadoffset=1;
    if (strstr(str_token,"footsteps")) isfootsteps=1;

    // ak je najdene cislo, naplni sa jednen zo 4 ciselnych udajov
    // patriacich animacii
    if (isNum(str_token) && !issex && !isheadoffset && !isfootsteps) {
     if (t==0) { anim[a].firstF=atoi(str_token); t++; }
     else if (t==1) { anim[a].numF=atoi(str_token); t++; }
          else if (t==2) { anim[a].loopingF=atoi(str_token); t++; }
               else if (t==3) { anim[a].fps=atoi(str_token); t++; }
     //printf("animacie %i %i %i %i %i\n",a,anim[a].firstF,anim[a].numF,anim[a].loopingF,anim[a].fps);
     if (t==4) a++;
    }
   }
  }
 }

 // v subore upper.md3 su definovane pohyby 0-4 + nasleduju pohyby pre TORSO
 // v subore lower.md3 su definovane pohyby 0-4 + nasleduju pohyby pre LEGS
 // preto je potrebne pre pohyby noh posunut prvy snimok na rovnaku poziciu
 // aku ma prvy snimok pohybu TORSO_GESTURE
 skip=anim[LEGS_WALKCR].firstF-anim[TORSO_GESTURE].firstF;
 for (a=LEGS_WALKCR;a<MAX_ANIMATIONS;a++) anim[a].firstF-=skip;

 // uprava poctu snimkov
 for (a=0;a<MAX_ANIMATIONS;a++) {
  if (anim[a].numF>0) anim[a].numF-=1;
  // INFO: prvy snimok daneho druhu pohybu
  //fprintf(stderr,"anim[%u]=%u\n",a,anim[a].firstF);
 }

 free(buffer);
 // zatvorenie suboru
 fclose(f);

 return(0);
}

// nastavenie parametrov animacie ******************************************
void CPlayer::setAnim(int L,int U)
{
 // animacia pre cast lower
 if (L!=-1) {
  if (act_l_anim!=L) { // ak sa ma zmenit animacia
   act_l_anim=L;
   anim_l=L;
   // first frame
   l_startFrame=anim[anim_l].firstF;
   // nextframe
   l_nextFrame=anim[anim_l].firstF;
   // num frames
   l_endFrame=anim[anim_l].firstF+anim[anim_l].numF;
   // frames per second
   l_fps=(float) anim[anim_l].fps;
  }
 }
 // animacia pre cast upper
 if (U!=-1) {
  if (act_u_anim!=U) { // ak sa ma zmenit animacia
   act_u_anim=U;
   anim_u=U;
   u_startFrame=anim[anim_u].firstF;
   u_nextFrame=anim[anim_u].firstF;
   u_endFrame=anim[anim_u].firstF+anim[anim_u].numF;
   u_fps=(float) anim[anim_u].fps;
   // ak sa ma nastavit animacia "strelba", Avatar striela
   if (anim_u==TORSO_ATTACK) { stopaction=0; Mix_PlayChannel(SHOT_WAV,sounds[SHOT_WAV], 0); }
  }
 }
}

// vykreslenie MD3 postavy *************************************************
int CPlayer::draw(vec3 AP,vec3 A,vec3 FP,vec3 FV,int T)
{
 draw_angle=A; // natacanie hornej casti postavy
 draw_type=T;  // sposob vykreslenia postavy (iba transformacie / aj vykreslenie)

 // plynutie casu
 aTime=SDL_GetTicks()/1200.0f;
 // vypocet parametrov snimku pre cast lower
 l_curTime=aTime;
 if (l_curTime-l_oldTime>1.0f/l_fps) {
  l_curFrame=l_nextFrame;
  l_nextFrame++;
  if (l_nextFrame>l_endFrame) l_nextFrame=l_startFrame;
  l_oldTime=l_curTime;
 }
 // vypocet parametrov snimku pre cast upper
 u_curTime=aTime;
 if (u_curTime-u_oldTime>1.0f/u_fps) {
  u_curFrame=u_nextFrame;
  u_nextFrame++;
  if (u_nextFrame>u_endFrame) u_nextFrame=u_startFrame;
  u_oldTime=u_curTime;
 }
 // nastavenie parametrov snimku pre cast lower
 l.setFrameInfo(l_curFrame,l_nextFrame,l_fps*(l_curTime-l_oldTime));
 //printf("%u %u\n",l_curFrame,l_nextFrame);
 // nastavenie parametrov snimku pre cast upper
 u.setFrameInfo(u_curFrame,u_nextFrame,u_fps*(u_curTime-u_oldTime));
 //printf("%u %u\n",u_curFrame,u_nextFrame);

 // po skonceni animacii "strelba" a "salutovanie" je postava "v klude"
 if (((anim_u==TORSO_ATTACK) || (anim_u==TORSO_GESTURE)) &&
     (u_nextFrame==u_endFrame)) setAnim(-1,TORSO_STAND);
 // po skonceni animacie "umierania" je postava "mrtva"
 if ((anim_u==BOTH_DEATH1) && (u_nextFrame==u_endFrame)) setAnim(BOTH_DEAD1,BOTH_DEAD1);

 // dopredu smerujuci vektor FV je nastaveny iba ak sa vykresluje protihrac
 // inak FV.x=2.0f
 if (FV.x!=2.0f) {
  EV=FP;           // poloha ustia hlavne
  TV=EV+FV*wRange; // bod na pozicii maximalneho dostrelu
/*
  // INFO: dopredu smerujuci vektor vo WORLD space
  glPushMatrix(); glLoadIdentity();
  glDisable(GL_TEXTURE_2D);
  glColor4f(0.0f,1.0f,0.0f,1.0f);
  glLineWidth(2.0f);
  glBegin(GL_LINES);
   glVertex3f(-EV.x,-EV.y,-EV.z); glVertex3f(-TV.x,-TV.y,-TV.z);
  glEnd();
  glLineWidth(1.0f);
  glColor4f(1.0f,1.0f,1.0f,1.0f);
  glEnable(GL_TEXTURE_2D);
  glPopMatrix();
*/
 }
 else TV.x=2.0f;

 impact=false; // indikator zasahu protihraca

 glPushMatrix();
  // tato cast je naprogramovana "na tvrdo" co nie je uplne dokonale,
  // plne vsak postacuje na ukazku jedneho z moznych sposobov realizacie
  // pouzite hodnoty su prisposobene konkretnej MD3 postave, v pripade
  // inej postavy bude potrebne udaje patricne upravit
  // mozno to riesit napr. externym suborom obsahujucim hodnoty pre rozne druhy postav,
  // alebo efektivnejsie ziskanim suradnic konkretneho bodu patriacemu niektoremu
  // z trojuholnikov, z ktorych je postava zlozena - napr. z hlavy (head) postavy
  switch (draw_type) {
   case 0: break; // iba transformacie, bez vykreslenia objektov
   case 1:
   case 2: glTranslatef(-AP.x,-AP.y+0.2f,-AP.z); // transformacie aj vykreslenie objektov
           // rotacia hornej casti postavy okolo osi Y
           glRotatef(-draw_angle.y,0.0f,1.0f,0.0f); break;
  }
  glPushMatrix();
   // vykreslenie vsetkych casti postavy - rekurzivna funkcia !!!
   drawPlayer(&l); // zacne sa castou lower
  glPopMatrix();
 glPopMatrix();

 if (impact) { // protihrac bol zasiahnuty gulkou
/*
  // INFO: zasiahnuty trojuholnik patriaci MD3 postave protihraca
  glPushMatrix(); glLoadIdentity();
  glDisable(GL_TEXTURE_2D);
  glColor4f(0.0f,1.0f,0.0f,1.0f);
  glLineWidth(2.0f);
  glBegin(GL_TRIANGLES);
   glVertex3fv(AA.get_data()); glVertex3fv(BB.get_data()); glVertex3fv(CC.get_data());
  glEnd();
  glLineWidth(1.0f);
  glColor4f(1.0f,1.0f,1.0f,1.0f);
  glEnable(GL_TEXTURE_2D);
  glPopMatrix();
*/
  return(1);
 }

 return(0);
}

// vykreslenie vsetkych casti postavy - rekurzivna funkcia !!! *************
void CPlayer::drawPlayer(CLoadMD3 *M)
{
 int i;

 // vykreslenie modelu
 switch (draw_type) {
  case 0: break;                                    // nic sa nekresli
  case 1: M->draw(); break;                         // kresli sa vsetko
  case 2: if ((M==&l) || (M==&w)) M->draw(); break; // kresli sa iba lower a weapon
 }

 // vsetky casti postavy, resp. vsetky modely priradene
 // znackam aktualneho modelu
 for (i=0;i<M->header.numTags;i++) {
  link=M->links[i];

  if (link) {
   // interpolacia
   p=M->ipol;

   // poloha
   pos=M->tags[M->actframe*M->header.numTags+i].origin;
   // natocenie
   rot=&M->tags[M->actframe*M->header.numTags+i].axis[0][0];
   // buduca poloha
   next_pos=M->tags[M->nextframe*M->header.numTags+i].origin;
   // buduce natocenie
   next_rot=&M->tags[M->nextframe*M->header.numTags+i].axis[0][0];

   // naplnenie matice (rotacia + posun)
   m[0]=-(rot[3]+p*(next_rot[3]-rot[3])); m[4]=(rot[6]+p*(next_rot[6]-rot[6])); m[8] =-(rot[0]+p*(next_rot[0]-rot[0]));
   m[1]=-(rot[4]+p*(next_rot[4]-rot[4])); m[5]=(rot[7]+p*(next_rot[7]-rot[7])); m[9] =-(rot[1]+p*(next_rot[1]-rot[1]));
   m[2]=-(rot[5]+p*(next_rot[5]-rot[5])); m[6]=(rot[8]+p*(next_rot[8]-rot[8])); m[10]=-(rot[2]+p*(next_rot[2]-rot[2]));
   m[3]=0.0f;                             m[7]=0.0f;                            m[11]=0.0f;
   m[12]=pos[0]+p*(next_pos[0]-pos[0]);
   m[13]=pos[1]+p*(next_pos[1]-pos[1]);
   m[14]=pos[2]+p*(next_pos[2]-pos[2]);
   m[15]=1.0f;

   glPushMatrix();

    // uprava OpenGL MODELVIEW matice
    glMultMatrixf(m);

    switch (draw_type) {
     case 0: if (link==&u) {
              // rotacia hornej casti (upper) postavy
              glRotatef(-draw_angle.y,0.0f,1.0f,0.0f);
              glRotatef(-draw_angle.x,1.0f,0.0f,0.0f);
              // nastavenie matice avamat - pre vypocet:
              // polohy ustia hlavne a dopredu smerujuceho vektora
              glGetFloatv(GL_MODELVIEW_MATRIX,avamat.get());
              avamat[12]*=-1.0f; avamat[13]*=-1.0f; avamat[14]*=-1.0f;
              if (ava==0) { // vojak
               // vypocet polohy ustia hlavne
               avamat[12]+=-0.3f*avamat[0]-0.8f*avamat[4]+0.5f*avamat[8];
               avamat[13]+=-0.3f*avamat[1]-0.8f*avamat[5]+0.5f*avamat[9];
               avamat[14]+=-0.3f*avamat[2]-0.8f*avamat[6]+0.5f*avamat[10];
               fireP.set(avamat[12],avamat[13],avamat[14]);
               // vypocet polohy vzdialeneho konca dopredu smerujuceho vektora
               pV=fireP;
               pV.x+=-0.14f*avamat[0]+0.11f*avamat[4]+1.0f*avamat[8];
               pV.y+=-0.14f*avamat[1]+0.11f*avamat[5]+1.0f*avamat[9];
               pV.z+=-0.14f*avamat[2]+0.11f*avamat[6]+1.0f*avamat[10];
               // dopredu smerujuci vektor
               fireV=pV-fireP; fireV.normalize();
		      }
		      else { // netvor
			   // vypocet polohy ustia hlavne
               avamat[12]+=-0.4f*avamat[0]-0.6f*avamat[4]+0.5f*avamat[8];
               avamat[13]+=-0.4f*avamat[1]-0.6f*avamat[5]+0.5f*avamat[9];
               avamat[14]+=-0.4f*avamat[2]-0.6f*avamat[6]+0.5f*avamat[10];
               fireP.set(avamat[12],avamat[13],avamat[14]);
               // vypocet polohy vzdialeneho konca dopredu smerujuceho vektora
               pV=fireP;
               pV.x+=0.04f*avamat[0]+0.1f*avamat[4]+1.0f*avamat[8];
               pV.y+=0.04f*avamat[1]+0.1f*avamat[5]+1.0f*avamat[9];
               pV.z+=0.04f*avamat[2]+0.1f*avamat[6]+1.0f*avamat[10];
               // dopredu smerujuci vektor
               fireV=pV-fireP; fireV.normalize();
			  }
             }
             if (link==&h) { // ak sa kresli hlava (head) postavy
              // nastavenie matice avamat
              glGetFloatv(GL_MODELVIEW_MATRIX,avamat.get());
              avamat[12]*=-1.0f; avamat[13]*=-1.0f; avamat[14]*=-1.0f;

              avamat.transpose(avamat);
              // posun polohy oci Avatara voci MD3 postave (v interface.cpp)
              avamat[12]+=0.05f*avamat[0]-0.35f*avamat[1]+0.05f*avamat[2];
              avamat[13]+=0.05f*avamat[4]-0.35f*avamat[5]+0.05f*avamat[6];
              avamat[14]+=0.05f*avamat[8]-0.35f*avamat[9]+0.05f*avamat[10];
             }
             break;
     case 1:
     case 2: if (link==&u) {
              // rotacia hornej (upper) casti postavy okolo osi X
              glRotatef(-draw_angle.x,1.0f,0.0f,0.0f);
             }
             break;
    }

    // rekurzivne vykreslenie jednotlivych casti postavy
    drawPlayer(link);

   glPopMatrix();
  }
 }
}

// vytvorenie odtlackov ****************************************************
int CPlayer::fire(vec3 EP,vec3 TP,vec3 RP,GLubyte F)
{
 if (stopaction==1) return(0); // Avatar moze strielat iba impulzne
                               // - synchronne s animaciou "strelba"

 // pridanie odtlacku (decal) --------------------------------------------

 // normalizovane vektory smerujuce dopredu, doprava a hore
 frontV=TP; rightV=RP; upV=frontV.cross(rightV);

 ip0=wRange;      // maximalny dostrel
 TP=EP+TP*wRange; // bod na pozicii maximalneho dostrelu

 // pre vsetky trojuholniky
 loopvk(tris) {
  // ak luc zasiahol trojuholnik
  if (m3d.ray_tri_collision(-EP,-TP,tris[k].n,tris[k].d,tris[k].cn[0],tris[k].cn[1],tris[k].cn[2])) {
   pV=m3d.get_ipoint(); // priesecnik luca a trojuholnika
   ip1=(pV+EP).magnitude();
   if (ip1<ip0) { // vyber najblizsieho priesecnika
    ipoint=pV; inormal=tris[k].n; ip0=ip1;
   }
  }
 }

 // ak avatar zasiahol nejaky trojuholnik
 if (ip0!=wRange) {
  // priprava 6-tich orezavacich rovin (steny 3D hranola)
  cplane[0].x=rightV.x; cplane[0].y=rightV.y; cplane[0].z=rightV.z;    // prava stena
  pV=ipoint+rightV*wDecalHalfSize;
  cplane[0].w=-(cplane[0].x*pV.x+cplane[0].y*pV.y+cplane[0].z*pV.z);

  cplane[1].x=-rightV.x; cplane[1].y=-rightV.y; cplane[1].z=-rightV.z; // lava stena
  pV=ipoint-rightV*wDecalHalfSize;
  cplane[1].w=-(cplane[1].x*pV.x+cplane[1].y*pV.y+cplane[1].z*pV.z);

  cplane[2].x=upV.x; cplane[2].y=upV.y; cplane[2].z=upV.z;             // vrchna stena
  pV=ipoint+upV*wDecalHalfSize;
  cplane[2].w=-(cplane[2].x*pV.x+cplane[2].y*pV.y+cplane[2].z*pV.z);

  cplane[3].x=-upV.x; cplane[3].y=-upV.y; cplane[3].z=-upV.z;          // spodna stena
  pV=ipoint-upV*wDecalHalfSize;
  cplane[3].w=-(cplane[3].x*pV.x+cplane[3].y*pV.y+cplane[3].z*pV.z);

  cplane[4].x=frontV.x; cplane[4].y=frontV.y; cplane[4].z=frontV.z;    // zadna stena
  pV=ipoint+frontV*wDecalHalfSize;
  cplane[4].w=-(cplane[4].x*pV.x+cplane[4].y*pV.y+cplane[4].z*pV.z);

  cplane[5].x=-frontV.x; cplane[5].y=-frontV.y; cplane[5].z=-frontV.z; // predna stena
  pV=ipoint-frontV*wDecalHalfSize;
  cplane[5].w=-(cplane[5].x*pV.x+cplane[5].y*pV.y+cplane[5].z*pV.z);

  addDecal(cplane); // pridanie odtlacku (vsetkych casti odtlacku)
 } // ak avatar zasiahol nejaky trojuholnik

 return(0);
}

// pridanie odtlacku (vsetkych casti odtlacku) *****************************
int CPlayer::addDecal(vec4 *P)
{
 // oznacenie casti posledneho pridaneho odtlacku
 lastG=NULL;

 // pre vsetky trojuholniky
 loopvk(tris) {
  // ak trojuholnik zasahuje do 3D hranola
  if (m3d.tri_in_box(tris[k].v[0],tris[k].v[1],tris[k].v[2],P)) {

   if (firstD==NULL) { // vyhradenie pamate pre 1. odtlacok
    firstD=new wDecal;
    if (firstD==NULL) {
     printf("Error allocating memory for weapon decal!\n"); return(1);
    }
    lastD=firstD;
   }
   else {              // vyhradenie pamate pre dalsi odtlacok
    lastD->next=new wDecal;
    if (lastD->next==NULL) {
     printf("Error allocating memory for weapon decal!\n"); return(1);
    }
    lastD=lastD->next;
   }

   // nasledujuci odtlacok
   lastD->next=NULL;

   // vrcholy casti odtlacku = vrcholy trojuholnika
   lastD->v[0]=tris[k].v[0]; lastD->v[1]=tris[k].v[1]; lastD->v[2]=tris[k].v[2];
   // normala = normala trojuholnika
   lastD->n=tris[k].n;
   // identifikator zavisly na normale, ak je normala casti odtlacku
   // postacujuco rovnobezna s inormal, potom idn=1 => tato cast
   // odtlacku lezi v rovine zasiahnutej gulkou
   d1=fabs(lastD->n.dot(inormal));
   if (d1>0.80f) lastD->idn=1; else lastD->idn=0;
   // orezanie casti odtlacku pomocou 6-tich stien 3D hranola
   lastD->vnum=m3d.tri_box_clip(lastD->v,P);

   // ak ma cast odtlacku po orezani aspon 3 vrcholy
   if (lastD->vnum>=3) {
    // ak este neboli oznacene casti posledneho pridaneho odtlacku
    if (lastG==NULL) lastG=lastD;
    // priesvitnost
    lastD->alpha=1.0f;
    // 8 vrcholov 3D hranola
    //lastD->abcd[0]=ipoint-rightV*wDecalHalfSize+frontV*wDecalHalfSize-upV*wDecalHalfSize;
    //lastD->abcd[1]=ipoint+rightV*wDecalHalfSize+frontV*wDecalHalfSize-upV*wDecalHalfSize;
    //lastD->abcd[2]=ipoint+rightV*wDecalHalfSize+frontV*wDecalHalfSize+upV*wDecalHalfSize;
    //lastD->abcd[3]=ipoint-rightV*wDecalHalfSize+frontV*wDecalHalfSize+upV*wDecalHalfSize;
    //lastD->abcd[4]=ipoint-rightV*wDecalHalfSize-frontV*wDecalHalfSize-upV*wDecalHalfSize;
    //lastD->abcd[5]=ipoint+rightV*wDecalHalfSize-frontV*wDecalHalfSize-upV*wDecalHalfSize;
    //lastD->abcd[6]=ipoint+rightV*wDecalHalfSize-frontV*wDecalHalfSize+upV*wDecalHalfSize;
    //lastD->abcd[7]=ipoint-rightV*wDecalHalfSize-frontV*wDecalHalfSize+upV*wDecalHalfSize;
   }
   else delDecal(lastD); // ak nema cast odtlacku min. 3 vrcholy, mozno ju vymazat

  } // ak trojuholnik zasahuje do 3D hranola
 } // pre vsetky trojuholniky

 // ak neboli oznacene casti posledneho pridaneho odtlacku => nebol pridany novy
 // odtlacok a teda nemusi prebehnut vypocet texturovacich suradnic
 if (lastG==NULL) return(1);

 // vypocet texturovacich suradnic -----------------------------------------

 // uprava velkosti odtlacku pre texturovanie
 ip0=1.0f/(wDecalHalfSize*4.0f); ip1=1.0f/(wDecalHalfSize*4.0f);
 // nahodny vyber 1 zo 4 textur, ktore reprezentuju odtlacok
 if (rand()<(RAND_MAX/2)) texU=0.25f; else texU=0.75f;
 if (rand()<(RAND_MAX/2)) texV=0.25f; else texV=0.75f;
 // vypocet uhlov pre otocenie vrcholov casti odtlacku do spolocnej roviny,
 // v ktorej prebehne vypocet texturovacich suradnic
 angR=PI2-inormal.angle(rightV); angU=PI2-inormal.angle(upV);

 // pre vsetky casti posledneho pridaneho odtlacku
 curD=lastG;
 while (curD!=NULL) {

  ind=0; // pociatocna hodnota indikatora (hrana, vrchol, neotacat)

  // ak normala casti odtlacku != inormala
  if (curD->idn==0) {
   // aktualna cast posledneho pridaneho odtlacku sa porovna s castami,
   // ktore lezia v rovine zasiahnutej gulkou
   cD=lastG;
   while (cD!=NULL) {
    // ak ide o rovnake casti, alebo ak porovnavana cast nelezi v rovine zasiahnutej gulkou
    if ((cD==curD)||(cD->idn==0)) { cD=cD->next; continue; }

    // najdenie spolocnej hrany 
    loopj(curD->vnum-1) { // pre vsetky hrany casti odtlacku
     loopi(cD->vnum-1) {  // pre vsetky hrany porovnavanej casti odtlacku
      // vypocet vzdialenosti vrcholov
      u1=(curD->v[j]-cD->v[i]);     // 1. vrchol - 1. vrchol
      u2=(curD->v[j+1]-cD->v[i+1]); // 2. vrchol - 2. vrchol
      u3=(curD->v[j]-cD->v[i+1]);   // 1. vrchol - 2. vrchol
      u4=(curD->v[j+1]-cD->v[i]);   // 2. vrchol - 1. vrchol
      d1=u1.magnitude(); d2=u2.magnitude(); d3=u3.magnitude(); d4=u4.magnitude();
      // ak su vrcholy hran postacujuco blizko, bola najdena spolocna hrana
      if (((d1<0.02f)&&(d2<0.02f))||((d3<0.02f)&&(d4<0.02f))) {
       u1=curD->v[j+1]-curD->v[j]; // hrana
       u2=curD->v[j]+u1*0.5f;      // vztazny bod
       u1.normalize();
       ind=1; break; // bola najdena spolocna hrana
      }
     }
     if (ind!=0) break; // ukoncenie cyklu loopj ak ind!=0
    }
    if (ind!=0) break; // ukoncenie cyklu while

    // najdenie spolocneho vrcholu
    loopj(curD->vnum) { // pre vsetky vrcholy casti odtlacku
     loopi(cD->vnum) {  // pre vsetky vrcholy porovnavanej casti odtlacku
      // vypocet vzdialenosti vrcholov
      u1=(curD->v[j]-cD->v[i]);
      d1=u1.magnitude();
      // ak su vrcholy postacujuco blizko, bol najdeny spolocny vrchol
      if (d1<0.02f) {
       // vrcholy hrany porovnavanej casti odtlacku
       u1=cD->v[i]; if ((i+1)<cD->vnum) u2=cD->v[i+1]; else u2=cD->v[0];
       u1=u1-u2;             // 1. vektor pre otocenie
       u1.normalize();
       u2=inormal.cross(u1); // 2. vektor pre otocenie
       u3=cD->v[i];          // vztazny bod
       ind=2; break; // bol najdeny spolocny vrchol
      }
     }
     if (ind!=0) break; // ukoncenie cyklu loopj ak ind!=0
    }
    if (ind!=0) break; // ukoncenie cyklu while

    cD=cD->next;
   }
   // ak nebola najdena spolocna hrana ani spolocny vrchol, cast odtlacku
   // sa nebude nijako otacat a priamo sa otexturuje, to sposobi vznik
   // nepeknych casti odtlackov, tuto nedokonalost vsak mozme zanedbat
   if (ind==0) ind=3;
  }

  if (ind==1) { // ak bola najdena spolocna hrana
   // vypocet uhla pre otocenie vrcholov okolo spolocnej hrany
   d3=inormal.angle(curD->n);
   // pre vsetky vrcholy casti odtlacku
   loopk(curD->vnum) {
    // posun vrcholu voci vztaznemu bodu
    pV=curD->v[k]-u2;
    // otacat sa bude v kladnom / zapornom smere
    d2=pV.x*inormal.x+pV.y*inormal.y+pV.z*inormal.z;
    if (d2>=0.0f) pV.rotate(-d3,u1); else pV.rotate(d3,u1);
    // posun vrcholu na povodnu poziciu a nasledny posun voci ipoint
    pV=(pV+u2)-ipoint;
    // otocenie vrcholu do spolocnej roviny
    pV.rotate(angU,rightV); pV.rotate(-angR,upV);
    // vypocet texturovacich suradnic
    d1=pV.dot(rightV)*ip0+texU; d2=pV.dot(upV)*ip1+texV;
    // orezanie texturovacich suradnic
    d1=max(texU-0.25f,min(d1,texU+0.25f)); d2=max(texV-0.25f,min(d2,texV+0.25f));
    // priradenie texturovacich suradnic konkretnemu vrcholu casti odtlacku
    curD->t[k].set(d1,d2);
    //curD->abcd[k]=ipoint+pV;
   }
  }
  else if (ind==2) { // ak bol najdeny spolocny vrchol
        // vypocet uhlov pre otocenie vrcholov okolo spolocneho vrcholu
        d3=PI2-curD->n.angle(u1); d4=PI2-curD->n.angle(u2);
        // pre vsetky vrcholy casti odtlacku
        loopk(curD->vnum) {
         // posun vrcholu voci vztaznemu bodu
         pV=curD->v[k]-u3;
         pV.rotate(d4,u1); pV.rotate(-d3,u2);
         // posun vrcholu na povodnu poziciu a nasledny posun voci ipoint
         pV=(pV+u3)-ipoint;
         // otocenie vrcholu do spolocnej roviny
         pV.rotate(angU,rightV); pV.rotate(-angR,upV);
         // vypocet texturovacich suradnic
         d1=pV.dot(rightV)*ip0+texU; d2=pV.dot(upV)*ip1+texV;
         // orezanie texturovacich suradnic
         d1=max(texU-0.25f,min(d1,texU+0.25f)); d2=max(texV-0.25f,min(d2,texV+0.25f));
         // priradenie texturovacich suradnic konkretnemu vrcholu casti odtlacku
         curD->t[k].set(d1,d2);
         //curD->abcd[k]=ipoint+pV;
        }
       }
       else {
        // pre vsetky vrcholy casti odtlacku
        loopk(curD->vnum) {
         // posun vrcholu voci ipoint
         pV=curD->v[k]-ipoint;
         // ak cast odtlacku lezi v rovine zasiahnutej gulkou, jej vrcholy
         // sa otocia do spolocnej roviny, ak cast odtlacku nelezi v rovine
         // zasiahnutej gulkou a zaroven nebola najdena spolocna hrana ani
         // vrchol, cast odtlacku sa otexturuje bez otacania
         if (ind==0) {
          // otocenie vrcholu
          pV.rotate(angU,rightV); pV.rotate(-angR,upV);
         }
         // vypocet texturovacich suradnic
         d1=pV.dot(rightV)*ip0+texU; d2=pV.dot(upV)*ip1+texV;
         // orezanie texturovacich suradnic
         d1=max(texU-0.25f,min(d1,texU+0.25f)); d2=max(texV-0.25f,min(d2,texV+0.25f));
         // priradenie texturovacich suradnic konkretnemu vrcholu casti odtlacku
         curD->t[k].set(d1,d2);
         //curD->abcd[k]=ipoint+pV;
        }
       }

  curD=curD->next;
 } // pre vsetky casti posledneho pridaneho odtlacku

 return(0);
}

// vykreslenie dier po gulkach *********************************************
void CPlayer::drawDecals(void)
{
 // vypnutie zapisu do DepthBuffera
 glDepthMask(GL_FALSE);
 // posun Z-hodnot
 glEnable(GL_POLYGON_OFFSET_FILL);
 // priesvitnost odtlacku
 glEnable(GL_BLEND);
 glBlendFunc(GL_DST_COLOR,GL_SRC_COLOR);
 // textura odtlacku
 glBindTexture(GL_TEXTURE_2D,texture.tex_id[tid[4]]);
 // shadery
 decal_vp.PRGenable(); decal_vp.PRGbind();
 decal_fp.PRGenable(); decal_fp.PRGbind();

 // pouziju sa polia atributov vrcholov:
 // 0 - suradnice vrcholu
 // 8 - texturovacie suradnice
 glEnableVertexAttribArray_(0);
 glEnableVertexAttribArray_(8);

 // pre vsetky odtlacky (pre vsetky casti odtlackov)
 curD=firstD;
 while (curD!=NULL) {
  glColor4f(0.5f,0.5f,0.5f,curD->alpha);            // farba
  glVertexAttribPointer_(0,3,GL_FLOAT,0,0,curD->v); // vrcholy
  glVertexAttribPointer_(8,2,GL_FLOAT,0,0,curD->t); // texturovacie suradnice
  glDrawArrays(GL_TRIANGLE_FAN,0,curD->vnum);
/*
  decal_fp.PRGdisable(); decal_vp.PRGdisable();

  glColor4f(1.0f,1.0f,0.0f,1.0f);
  glBegin(GL_LINE_LOOP);
   loopl(curD->vnum) glVertex3fv(curD->abcd[l].v);
  glEnd();

  decal_vp.PRGenable(); decal_vp.PRGbind();
  decal_fp.PRGenable(); decal_fp.PRGbind();
*/
  // priesvitnost odtlacku
  curD->alpha-=0.004f; if (curD->alpha<0.0f) delDecal(curD);

  if (curD!=NULL) curD=curD->next;
 }

 // standardna farba
 glColor4f(1.0f,1.0f,1.0f,1.0f);

 glDisableVertexAttribArray_(0);
 glDisableVertexAttribArray_(8);

 // vypnutie shaderov
 decal_fp.PRGdisable(); decal_vp.PRGdisable();
 // vypnutie priesvitnosti
 glDisable(GL_BLEND);
 // vypnutie posunu Z-hodnot
 glDisable(GL_POLYGON_OFFSET_FILL);
 // zapnutie zapisu do DepthBuffera
 glDepthMask(GL_TRUE);
}

// vymazanie odtlacku, alebo casti odtlacku ********************************
void CPlayer::delDecal(wDecal *P)
{
 if (P==firstD) { // zmazanie 1. odtlacku
  firstD=firstD->next;
  DELETEP(P);
  curD=firstD;
 }
 else {           // zmazanie odtlacku v zozname
  curD=firstD;
  while (curD->next!=P) curD=curD->next;
  curD->next=P->next;
  DELETEP(P);
  if (curD->next==NULL) lastD=curD;
 }
}

// zmazanie vsetkych odtlackov (vsetkych casti) ****************************
void CPlayer::delAllDecals(void)
{
 curD=firstD;
 if (curD!=NULL) {
  while (curD->next!=NULL) { // zmazanie vsetkych odtlackov zo zoznamu
   firstD=curD->next;
   DELETEP(curD);
   curD=firstD;
  }
  DELETEP(firstD);           // zmazanie 1. odtlacku
 }

 // prvy, posledny, aktualny odtlacok (casti odtlacku),
 // oznacenie casti posledneho pridaneho odtlacku
 firstD=NULL; lastD=NULL; curD=NULL; cD=NULL; lastG=NULL;
}
