// hlavickove subory *******************************************************
#include "../includes.h"
// vlastny .h subor
#include "octree.h"

// konstruktor *************************************************************
COctree::COctree()
{
 nodes_num=0; // pocet listov oktanoveho stromu
 tris_num=0;  // pocet odkazov na trojuholniky, ktore su ulozene v oktanovom strome

 view_tris_num=0; // pocet aktivnych ukazovatelov na viditelne trojuholniky
 tests_num=0;     // pocet testov na viditelnost listov oktanoveho stromu
 col_tris_num=0;  // pocet aktivnych ukazovatelov na kolizne trojuholniky

 alloc_mem=0; free_mem=0; // pocitadla vyhradenej/uvolnenej pamate

 otree=NULL; // rodicovsky list oktanoveho stromu
}

// destruktor **************************************************************
COctree::~COctree()
{
 destroy(); // zrusenie oktanoveho stromu
}

// vytvorenie oktanoveho stromu ********************************************
int COctree::create(void)
{
 // pocitadla vyhradenej/uvolnenej pamate
 alloc_mem=0; free_mem=0;

 otree=new _tree_node; // vytvorenie rodicovskeho listu oktanoveho stromu
 if (otree==NULL) {
  printf("Chyba vyhradenia pamate pre rodicovsky list oktanoveho stromu!\n"); return(1);
 }
 alloc_mem+=sizeof(_tree_node);

 // identifikator rodicovskeho listu oktanoveho stromu
 otree->id=nodes_num; nodes_num++;
 
 // pociatocne rozmery rodicovskeho listu oktanoveho stromu (AABB)
 otree->minx=999999.9f; otree->maxx=-999999.9f;
 otree->miny=999999.9f; otree->maxy=-999999.9f;
 otree->minz=999999.9f; otree->maxz=-999999.9f;
 // pre vsetky trojuholniky
 loopvk(tris) {
  // zarovnanie rozmerov rodivoskeho listu podla suradnic trojuholnikov
  if (tris[k].minmaxx.x<otree->minx) otree->minx=tris[k].minmaxx.x;
  if (tris[k].minmaxy.x<otree->miny) otree->miny=tris[k].minmaxy.x;
  if (tris[k].minmaxz.x<otree->minz) otree->minz=tris[k].minmaxz.x;
  if (tris[k].minmaxx.y>otree->maxx) otree->maxx=tris[k].minmaxx.y;
  if (tris[k].minmaxy.y>otree->maxy) otree->maxy=tris[k].minmaxy.y;
  if (tris[k].minmaxz.y>otree->maxz) otree->maxz=tris[k].minmaxz.y;
 }

 // rodicovsky list nema ziadne trojuholniky
 otree->node_tris=NULL;
 otree->node_tris_num=0;

 // vsetky listy (podlisty) budu zatial prazdne
 loopi(8) otree->node[i]=NULL;

 // nastavenie indikatora viditelnosti listu
 otree->view_bit=true;

 // pridanie listov do oktanoveho stromu - REKURZIVNA FUNKCIA
 tree_level=0; // aktualna hlbka oktanoveho stromu
 node_add(otree);

 // INFO: vyhradena pamat
 printf("Velkost pamate vyhradenej pre oktanovy strom: %u\n",alloc_mem);

 free_mem=0; // moze dojst k uvolneniu prazdnych listov

 // zaradenie trojuholnikov do listov oktanoveho stromu - REKURZVINA FUNKCIA
 tree_level=0; // aktualna hlbka oktanoveho stromu
 tris_add(otree);

 // INFO: uvolnena pamat z dovodu prazdnych listov
 printf("Velkost pamate uvolnenej prazdnymi listami: %u\n",free_mem);

 return(0);
}

// pridanie listov do oktanoveho stromu - REKURZIVNA FUNKCIA ***************
int COctree::node_add(_tree_node *N)
{
 // ak sa dosiahne maximalna hlbka, rekurzia pokracuje na predoslej hlbke
 if (tree_level==TREE_DEPTH) { tree_level--; return(0); }

 // 4 predne listy (kocky) -------------------------------------------------

 // lava dolna kocka
 N->node[0]=new _tree_node;
 if (N->node[0]==NULL) { printf("Chyba vyhradenia pamate pre list: node0!\n"); return(1); }
 alloc_mem+=sizeof(_tree_node);

 // nastavenie rozmerov kocky
 N->node[0]->minx=N->minx; N->node[0]->maxx=N->minx+(N->maxx-N->minx)/2.0f;
 N->node[0]->miny=N->miny; N->node[0]->maxy=N->miny+(N->maxy-N->miny)/2.0f;
 N->node[0]->minz=N->minz; N->node[0]->maxz=N->minz+(N->maxz-N->minz)/2.0f;
 // INFO: pridanie listu
 //printf("front left down\n");

 // identifikator listu
 N->node[0]->id=nodes_num; nodes_num++;
 N->node[0]->node_tris=NULL; // list zatial nema ziadne trojuholniky
 N->node[0]->node_tris_num=0;
 loopi(8) N->node[0]->node[i]=NULL; // vsetky listy (podlisty) budu zatial prazdne
 N->node[0]->view_bit=true; // nastavenie indikatora viditelnosti listu

 tree_level++; // nova hlbka
 // INFO: aktualna hlbka oktanoveho stromu
 //printf("tree_level: %u\n",tree_level);
 node_add(N->node[0]); // REKURZIA - leve dolne listy

 // prava dolna kocka
 N->node[1]=new _tree_node;
 if (N->node[1]==NULL) { printf("Chyba vyhradenia pamate pre list: node1!\n"); return(1); }
 alloc_mem+=sizeof(_tree_node);

 // nastavenie rozmerov kocky
 N->node[1]->minx=N->minx+(N->maxx-N->minx)/2.0f; N->node[1]->maxx=N->maxx;
 N->node[1]->miny=N->miny;                        N->node[1]->maxy=N->miny+(N->maxy-N->miny)/2.0f;
 N->node[1]->minz=N->minz;                        N->node[1]->maxz=N->minz+(N->maxz-N->minz)/2.0f;
 // INFO: pridanie listu
 //printf("front right down\n");

 // identifikator listu
 N->node[1]->id=nodes_num; nodes_num++;
 N->node[1]->node_tris=NULL; // list zatial nema ziadne trojuholniky
 N->node[1]->node_tris_num=0;
 loopi(8) N->node[1]->node[i]=NULL; // vsetky listy (podlisty) budu zatial prazdne
 N->node[1]->view_bit=true; // nastavenie indikatora viditelnosti listu

 tree_level++; // nova hlbka
 // INFO: aktualna hlbka oktanoveho stromu
 //printf("tree_level: %u\n",tree_level);
 node_add(N->node[1]); // REKURZIA - prave dolne listy

 // prava horna kocka
 N->node[2]=new _tree_node; 
 if (N->node[2]==NULL) { printf("Chyba vyhradenia pamate pre list: node2!\n"); return(1); }
 alloc_mem+=sizeof(_tree_node);

 // nastavenie rozmerov kocky
 N->node[2]->minx=N->minx+(N->maxx-N->minx)/2.0f; N->node[2]->maxx=N->maxx;
 N->node[2]->miny=N->miny+(N->maxy-N->miny)/2.0f; N->node[2]->maxy=N->maxy;
 N->node[2]->minz=N->minz;                        N->node[2]->maxz=N->minz+(N->maxz-N->minz)/2.0f;
 // INFO: pridanie listu
 //printf("front right up\n");

 // identifikator listu
 N->node[2]->id=nodes_num; nodes_num++;
 N->node[2]->node_tris=NULL; // list zatial nema ziadne trojuholniky
 N->node[2]->node_tris_num=0;
 loopi(8) N->node[2]->node[i]=NULL; // vsetky listy (podlisty) budu zatial prazdne
 N->node[2]->view_bit=true; // nastavenie indikatora viditelnosti listu

 tree_level++; // nova hlbka
 // INFO: aktualna hlbka oktanoveho stromu
 //printf("tree_level: %u\n",tree_level);
 node_add(N->node[2]); // REKURZIA - prave horne listy

 // lava horna kocka
 N->node[3]=new _tree_node;
 if (N->node[3]==NULL) { printf("Chyba vyhradenia pamate pre list: node3!\n"); return(1); }
 alloc_mem+=sizeof(_tree_node);

 // nastavenie rozmerov kocky
 N->node[3]->minx=N->minx;                        N->node[3]->maxx=N->minx+(N->maxx-N->minx)/2.0f;
 N->node[3]->miny=N->miny+(N->maxy-N->miny)/2.0f; N->node[3]->maxy=N->maxy;
 N->node[3]->minz=N->minz;                        N->node[3]->maxz=N->minz+(N->maxz-N->minz)/2.0f;
 // INFO: pridanie listu
 //printf("front left up\n");

 // identifikator listu
 N->node[3]->id=nodes_num; nodes_num++;
 N->node[3]->node_tris=NULL; // list zatial nema ziadne trojuholniky
 N->node[3]->node_tris_num=0;
 loopi(8) N->node[3]->node[i]=NULL; // vsetky listy (podlisty) budu zatial prazdne
 N->node[3]->view_bit=true; // nastavenie indikatora viditelnosti listu

 tree_level++; // nova hlbka
 // INFO: aktualna hlbka oktanoveho stromu
 //printf("tree_level: %u\n",tree_level);
 node_add(N->node[3]); // REKURZIA - lave horne listy

 // 4 zadne listy (kocky) --------------------------------------------------

 // lava dolna kocka
 N->node[4]=new _tree_node; 
 if (N->node[4]==NULL) { printf("Chyba vyhradenia pamate pre list: node4!\n"); return(1); }
 alloc_mem+=sizeof(_tree_node);

 // nastavenie rozmerov kocky
 N->node[4]->minx=N->minx;                        N->node[4]->maxx=N->minx+(N->maxx-N->minx)/2.0f;
 N->node[4]->miny=N->miny;                        N->node[4]->maxy=N->miny+(N->maxy-N->miny)/2.0f;
 N->node[4]->minz=N->minz+(N->maxz-N->minz)/2.0f; N->node[4]->maxz=N->maxz;
 // INFO: pridanie listu
 //printf("back left down\n");

 // identifikator listu
 N->node[4]->id=nodes_num; nodes_num++;
 N->node[4]->node_tris=NULL; // list zatial nema ziadne trojuholniky
 N->node[4]->node_tris_num=0;
 loopi(8) N->node[4]->node[i]=NULL; // vsetky listy (podlisty) budu zatial prazdne
 N->node[4]->view_bit=true; // nastavenie indikatora viditelnoti listu

 tree_level++; // nova hlbka
 // INFO: aktualna hlbka oktanoveho stromu
 //printf("tree_level: %u\n",tree_level);
 node_add(N->node[4]); // REKURZIA - lave dolne listy

 // prava dolna kocka
 N->node[5]=new _tree_node;
 if (N->node[5]==NULL) { printf("Chyba vyhradenia pamate pre list: node5!\n"); return(1); }
 alloc_mem+=sizeof(_tree_node);

 // nastavenie rozmerov kocky
 N->node[5]->minx=N->minx+(N->maxx-N->minx)/2.0f; N->node[5]->maxx=N->maxx;
 N->node[5]->miny=N->miny;                        N->node[5]->maxy=N->miny+(N->maxy-N->miny)/2.0f;
 N->node[5]->minz=N->minz+(N->maxz-N->minz)/2.0f; N->node[5]->maxz=N->maxz;
 // INFO: pridanie listu
 //printf("front right down\n");

 // identifikator listu
 N->node[5]->id=nodes_num; nodes_num++;
 N->node[5]->node_tris=NULL; // list zatial nema ziadne trojuholniky
 N->node[5]->node_tris_num=0;
 loopi(8) N->node[5]->node[i]=NULL; // vsetky listy (podlisty) budu zatial prazdne
 N->node[5]->view_bit=true; // nastavenie indikatora viditelnosti listu

 tree_level++; // nova hlbka
 // INFO: aktualna hlbka oktanoveho stromu
 //printf("tree_level: %u\n",tree_level);
 node_add(N->node[5]); // REKURZIA - prave dolne listy

 // prava horna kocka
 N->node[6]=new _tree_node;
 if (N->node[6]==NULL) { printf("Chyba vyhradenia pamate pre list: node6!\n"); return(1); }
 alloc_mem+=sizeof(_tree_node);

 // nastavenie rozmerov kocky
 N->node[6]->minx=N->minx+(N->maxx-N->minx)/2.0f; N->node[6]->maxx=N->maxx;
 N->node[6]->miny=N->miny+(N->maxy-N->miny)/2.0f; N->node[6]->maxy=N->maxy;
 N->node[6]->minz=N->minz+(N->maxz-N->minz)/2.0f; N->node[6]->maxz=N->maxz;
 // INFO: pridanie listu
 //printf("front right up\n");

 // identifikator listu
 N->node[6]->id=nodes_num; nodes_num++;
 N->node[6]->node_tris=NULL; // list zatial nema ziadne trojuholniky
 N->node[6]->node_tris_num=0;
 loopi(8) N->node[6]->node[i]=NULL; // vsetky listy (podlisty) budu zatial prazdne
 N->node[6]->view_bit=true; // nastavenie indikatora viditelnosti listu

 tree_level++; // nova hlbka
 // INFO: aktualna hlbka oktanoveho stromu
 //printf("tree_level: %u\n",tree_level);
 node_add(N->node[6]); // REKURZIA - prave horne listy

 // lava horna kocka
 N->node[7]=new _tree_node;
 if (N->node[7]==NULL) { printf("Chyba vyhradenia pamate pre list: node7!\n"); return(1); }
 alloc_mem+=sizeof(_tree_node);

 // nastavenie rozmerov kocky
 N->node[7]->minx=N->minx;                        N->node[7]->maxx=N->minx+(N->maxx-N->minx)/2.0f;
 N->node[7]->miny=N->miny+(N->maxy-N->miny)/2.0f; N->node[7]->maxy=N->maxy;
 N->node[7]->minz=N->minz+(N->maxz-N->minz)/2.0f; N->node[7]->maxz=N->maxz;
 // INFO: pridanie listu
 //printf("front left up\n");

 // identifikator listu
 N->node[7]->id=nodes_num; nodes_num++;
 N->node[7]->node_tris=NULL; // list zatial nema ziadne trojuholniky
 N->node[7]->node_tris_num=0;
 loopi(8) N->node[7]->node[i]=NULL; // vsetky listy (podlisty) budu zatial prazdne
 N->node[7]->view_bit=true; // nastavenie indikatora viditelnosti listu

 tree_level++; // nova hlbka
 // INFO: aktualna hlbka oktanoveho stromu
 //printf("tree_level: %u\n",tree_level);
 node_add(N->node[7]); // REKURZIA - lave horne listy

 // rekurzia pokracuje na predoslej hlbke
 tree_level--; return(0);
}

// zaradenie trojuholnikov do listov oktanoveho stromu - REKURZIVNA FUNKCIA
int COctree::tris_add(_tree_node *N)
{
 if (tree_level==TREE_DEPTH) { // pridava sa iba na najnizsiu hlbku
  // zistenie velkosti pamate, ktoru je potrebne vyhradit
  // pre pole odkazov na trojuholniky priradene listu
  loopvk(tris) { // pre vsetky trojuholniky
   // ak trojuholnik lezi mimo listu (kocky)
   if (tris[k].minmaxx.y<N->minx) continue;
   if (tris[k].minmaxx.x>N->maxx) continue;
   if (tris[k].minmaxy.y<N->miny) continue;
   if (tris[k].minmaxy.x>N->maxy) continue;
   if (tris[k].minmaxz.y<N->minz) continue;
   if (tris[k].minmaxz.x>N->maxz) continue;
   // trojuholnik lezi v liste (kocke) => zvysi sa pocet trojuholnikov listu
   N->node_tris_num++;
  }

  // INFO: pocet trojuholnikov listu
  //printf("node_tris_num: %u\n",N->node_tris_num);

  if (N->node_tris_num!=0) { // v liste su trojuholniky
   // INFO: identifikator listu s trojuholnikmi
   //printf(stderr,"%u ",N->id);

   // vyhradenie pamate pre pole odkazov na trojuholniky listu
   N->node_tris=(_tri **) malloc(N->node_tris_num*sizeof(_tri *));
   if (N->node_tris==NULL) {
    printf("Chyba vyhradenia pamate pre pole odkazov na trojuholniky listu!\n"); return(1);
   }

   // pocet odkazov na trojuholniky, ktore su ulozene v oktanovom strome
   tris_num+=N->node_tris_num;

   a=0;
   // naplnenie pola odkazov na trojuholniky priradene listu
   loopvk(tris) {
    // ak trojuholnik lezi mimo listu (kocky)
    if (tris[k].minmaxx.y<N->minx) continue;
    if (tris[k].minmaxx.x>N->maxx) continue;
    if (tris[k].minmaxy.y<N->miny) continue;
    if (tris[k].minmaxy.x>N->maxy) continue;
    if (tris[k].minmaxz.y<N->minz) continue;
    if (tris[k].minmaxz.x>N->maxz) continue;
    // trojuholnik lezi v liste (kocke) => odkaz na tento trojuholnik
    // sa umiestni do pola odkazov na trojuholniky priradene listu
    N->node_tris[a]=&tris[k]; a++;
   }
  }
  else { tree_level--; return(1); } // list je potrebne vymazat

  // rekurzia pokracuje na predoslej hlbke
  tree_level--; return(0);
 } // pridava sa iba na najnizsiu hlbku

 tree_level++;
 if (tris_add(N->node[0])) {
  // list sa vymaze, pretoze neobsahuje ziadne odkazy na trojuholniky
  if (N->node[0]->node_tris!=NULL) free(N->node[0]->node_tris); // zrusenie pola odkazov
  DELETEP(N->node[0]); // zrusenie listu
  free_mem+=sizeof(_tree_node); nodes_num--;
 }
 tree_level++;
 if (tris_add(N->node[1])) {
  // list sa vymaze, pretoze neobsahuje ziadne odkazy na trojuholniky
  if (N->node[1]->node_tris!=NULL) free(N->node[1]->node_tris);// zrusenie pola odkazov
  DELETEP(N->node[1]); // zrusenie listu
  free_mem+=sizeof(_tree_node); nodes_num--;
 }
 tree_level++;
 if (tris_add(N->node[2])) {
  // list sa vymaze, pretoze neobsahuje ziadne odkazy na trojuholniky
  if (N->node[2]->node_tris!=NULL) free(N->node[2]->node_tris);// zrusenie pola odkazov
  DELETEP(N->node[2]); // zrusenie listu
  free_mem+=sizeof(_tree_node); nodes_num--;
 }
 tree_level++;
 if (tris_add(N->node[3])) {
  // list sa vymaze, pretoze neobsahuje ziadne odkazy na trojuholniky
  if (N->node[3]->node_tris!=NULL) free(N->node[3]->node_tris);// zrusenie pola odkazov
  DELETEP(N->node[3]); // zrusenie listu
  free_mem+=sizeof(_tree_node); nodes_num--;
 }
 tree_level++;
 if (tris_add(N->node[4])) {
  // list sa vymaze, pretoze neobsahuje ziadne odkazy na trojuholniky
  if (N->node[4]->node_tris!=NULL) free(N->node[4]->node_tris);// zrusenie pola odkazov
  DELETEP(N->node[4]); // zrusenie listu
  free_mem+=sizeof(_tree_node); nodes_num--;
 }
 tree_level++;
 if (tris_add(N->node[5])) {
  // list sa vymaze, pretoze neobsahuje ziadne odkazy na trojuholniky
  if (N->node[5]->node_tris!=NULL) free(N->node[5]->node_tris);// zrusenie pola odkazov
  DELETEP(N->node[5]); // zrusenie listu
  free_mem+=sizeof(_tree_node); nodes_num--;
 }
 tree_level++;
 if (tris_add(N->node[6])) {
  // list sa vymaze, pretoze neobsahuje ziadne odkazy na trojuholniky
  if (N->node[6]->node_tris!=NULL) free(N->node[6]->node_tris);// zrusenie pola odkazov
  DELETEP(N->node[6]); // zrusenie listu
  free_mem+=sizeof(_tree_node); nodes_num--;
 }
 tree_level++;
 if (tris_add(N->node[7])) {
  // list sa vymaze, pretoze neobsahuje ziadne odkazy na trojuholniky
  if (N->node[7]->node_tris!=NULL) free(N->node[7]->node_tris);// zrusenie pola odkazov
  DELETEP(N->node[7]); // zrusenie listu
  free_mem+=sizeof(_tree_node); nodes_num--;
 }

 // rekurzia pokracuje na predoslej hlbke
 tree_level--; return(0);
}

// vykreslenie oktanoveho stromu *******************************************
void COctree::draw(void)
{
 glPushAttrib(GL_TEXTURE_BIT|GL_LIGHTING_BIT|GL_DEPTH_BUFFER_BIT);
 glDisable(GL_TEXTURE_2D); // vypnutie texturovania
 glDisable(GL_LIGHTING);   // vypnutie osvetlenia
 //glDisable(GL_DEPTH_TEST); // vypnutie testu hlbky

 glColor4f(1.0f,1.0f,0.0f,1.0f); // zlta farba

 // vykreslenie listov oktanoveho stromu - REKURZIVNA FUNKCIA
 tree_level=0;
 node_draw(otree);

 glColor4f(1.0f,1.0f,1.0f,1.0f); // biela farba

 glPopAttrib(); // vyber ulozenych atributov (TEXTURE,LIGHTING,DEPTH_BUFFER)
}

// vykreslenie listov oktanoveho stromu - REKURZIVNA FUNKCIA ***************
int COctree::node_draw(_tree_node *N)
{
 if (N!=NULL) {
  if ((tree_level==TREE_DEPTH)&&(N->view_bit)) { // ak je list (kocka) viditelny
   // predna stena
   glBegin(GL_LINE_LOOP);
    glVertex3f(N->minx,N->miny,N->minz); glVertex3f(N->maxx,N->miny,N->minz);
    glVertex3f(N->maxx,N->maxy,N->minz); glVertex3f(N->minx,N->maxy,N->minz);
   glEnd();
   // zadna stena
   glBegin(GL_LINE_LOOP);
    glVertex3f(N->minx,N->miny,N->maxz); glVertex3f(N->maxx,N->miny,N->maxz);
    glVertex3f(N->maxx,N->maxy,N->maxz); glVertex3f(N->minx,N->maxy,N->maxz);
   glEnd();
   // spojovacie ciary
   glBegin(GL_LINES);
    // lava dolna
    glVertex3f(N->minx,N->miny,N->minz); glVertex3f(N->minx,N->miny,N->maxz);
    // prava dolna
    glVertex3f(N->maxx,N->miny,N->minz); glVertex3f(N->maxx,N->miny,N->maxz);
    // prava horna
    glVertex3f(N->maxx,N->maxy,N->minz); glVertex3f(N->maxx,N->maxy,N->maxz);
    // lava horna
    glVertex3f(N->minx,N->maxy,N->minz); glVertex3f(N->minx,N->maxy,N->maxz);
   glEnd();
   
  }
 }

 // ak sa dosiahne maximalna hlbka, rekurzia pokracuje na predoslej hlbke
 if (tree_level==TREE_DEPTH) { tree_level--; return(0); }

 tree_level++; node_draw(N->node[0]);
 tree_level++; node_draw(N->node[1]);
 tree_level++; node_draw(N->node[2]);
 tree_level++; node_draw(N->node[3]);
 tree_level++; node_draw(N->node[4]);
 tree_level++; node_draw(N->node[5]);
 tree_level++; node_draw(N->node[6]);
 tree_level++; node_draw(N->node[7]);

 // rekurzia pokracuje na predoslej hlbke
 tree_level--; return(0);
}

// zrusenie oktanoveho stromu **********************************************
int COctree::destroy(void)
{
 // pocitadlo uvolnenej pamate
 free_mem=0;

 if (otree!=NULL) {

  // zrusenie listov oktanoveho stromu - REKURZIVNA FUNKCIA
  for (a=TREE_DEPTH;a>0;a--) {
   del_level=a;
   tree_level=0;
   node_del(otree);
  }

  // zrusenie pola odkazov na trojuholniky rodicovskeho listu
  if (otree->node_tris!=NULL) free(otree->node_tris);
  // zrusenie rodicovskeho listu
  DELETEP(otree);
  free_mem+=sizeof(_tree_node);
 }

 // INFO: uvolnena pamat
 printf("Velkost pamate uvolnenej oktanovym stromom: %u\n",free_mem);

 nodes_num=0; // pocet listov oktanoveho stromu
 tris_num=0;  // pocet odkazov na trojuholniky, ktore su ulozene v oktanovom strome

 return(0);
}

// zrusenie listov oktanoveho stromu - REKURZIVNA FUNKCIA ******************
int COctree::node_del(_tree_node *N)
{
 if ((tree_level==del_level)&&(N!=NULL)) { // rusi sa v konkretnej hlbke
  // zrusenie pola odkazov na trojuholniky
  if (N->node_tris!=NULL) free(N->node_tris);
  // zrusenie listu
  DELETEP(N);
  free_mem+=sizeof(_tree_node);
  // rekurzia pokracuje na predoslej hlbke
  tree_level--; return(0);
 }

 // ak sa dosiahne maximalna hlbka, rekurzia pokracuje na predoslej hlbke
 if (tree_level==TREE_DEPTH) { tree_level--; return(0); }

 tree_level++; node_del(N->node[0]);
 tree_level++; node_del(N->node[1]);
 tree_level++; node_del(N->node[2]);
 tree_level++; node_del(N->node[3]);
 tree_level++; node_del(N->node[4]);
 tree_level++; node_del(N->node[5]);
 tree_level++; node_del(N->node[6]);
 tree_level++; node_del(N->node[7]);

 // rekurzia pokracuje na predoslej hlbke
 tree_level--; return(0);
}

// vyber viditelnych a koliznych trojuholnikov *****************************
void COctree::select_tris(vec3 AP,GLfloat AE,GLfloat AW,bool S)
{
 // reset indikatora spracovania pre vsetky trojuholniky
 loopvk(tris) tris[k].done=false;

 tests_num=0; // pocet testov na viditelnost listov oktanoveho stromu

 if (!S) { // vyber viditelnych trojuholnikov
  view_tris_num=0;
  tree_level=0;
  select_view_tris(otree);
 }
 else { // vyber koliznych trojuholnikov
  ap=AP; ae=AE; aw=AW; // poloha Avatara, vyska oci a polovica sirky Avatara
  col_tris_num=0;
  tree_level=0;
  select_col_tris(otree);
 }
}

// vyber viditelnych trojuholnikov - REKURZIVNA FUNKCIA ********************
int COctree::select_view_tris(_tree_node *N)
{
 if (N!=NULL) {
  // pocet testov na viditelnost listov oktanoveho stromu
  tests_num++;

  // Frustum Culling listu oktanoveho stromu
  if (frustum.cube_in_frustum(N->minx,N->maxx,
                              N->miny,N->maxy,
                              N->minz,N->maxz)) {
   // list je viditelny
   N->view_bit=true;

   // vsetky odkazy na trojuholniky priradene listu
   // sa umiestnia do zoznamu ukazovatelov na viditelne trojuholniky
   // trojuholnikom sa nastavi indikator spracovania
   // pracuje sa v maximalnej hlbku stromu, pretoze iba v nej sa nachadzaju trojuholniky
   if (tree_level==TREE_DEPTH) {
    loopi(N->node_tris_num) {
     if (!(N->node_tris[i]->done)) {
      N->node_tris[i]->done=true; 
      view_tris[view_tris_num]=N->node_tris[i];
      view_tris_num++;
     }
    }
   }
  }
  else {
   // list nie je viditelny a ani jeden z jeho listov (podlistov)
   // nie je viditelny, takto sa jednoduchym testom odstrani velke mnozstvo trojuholnikov
   N->view_bit=false;

   // rekurzia pokracuje na predoslej hlbke
   tree_level--; return(0);
  }
 }

 // ak sa dosiahne maximalna hlbka, rekurzia pokracuje na predoslej hlbke
 if (tree_level==TREE_DEPTH) { tree_level--; return(0); }

 tree_level++; select_view_tris(N->node[0]);
 tree_level++; select_view_tris(N->node[1]);
 tree_level++; select_view_tris(N->node[2]);
 tree_level++; select_view_tris(N->node[3]);
 tree_level++; select_view_tris(N->node[4]);
 tree_level++; select_view_tris(N->node[5]);
 tree_level++; select_view_tris(N->node[6]);
 tree_level++; select_view_tris(N->node[7]);

 // rekurzia pokracuje na predoslej hlbke
 tree_level--; return(0);
}

// vyber koliznych trojuholnikov - REKURZIVNA FUNKCIA **********************
int COctree::select_col_tris(_tree_node *N)
{
 if (N!=NULL) {
  // vyberu su tzv. kolizne trojuholniky, s tymito
  // trojuholnikmi je potrebne vykonat test kolizie
  // testuje sa poloha Avatara zvacsena o polovicu jeho sirky
  // s listami (kockami) oktanoveho stromu
  // ak je list jednoznacne mimo rozmerov avatara,
  // pokracuje sa na predoslej hlbke cim sa jednoduchymi
  // testami odstrani velke mnostvo trojuholnikov
  if (ap.x-aw>N->maxx) { tree_level--; return(0); }
  else if (ap.x+aw<N->minx) { tree_level--; return(0); }
       else if (ap.y-ae>N->maxy) { tree_level--; return(0); }
            else if (ap.y+ae<N->miny) { tree_level--; return(0); }
                 else if (ap.z-aw>N->maxz) { tree_level--; return(0); }
                      else if (ap.z+aw<N->minz) { tree_level--; return(0); }
                           else {
                            // vsetky odkazy na trojuholniky priradene listu
                            // sa umiestnia do zoznamu ukazovatelov na kolizne trojuholniky
                            // trojuholnikom sa nastavi indikator spracovania
                            // pracuje sa v maximalnej hlbku stromu, pretoze iba v nej sa nachadzaju trojuholniky
                            if (tree_level==TREE_DEPTH) {
                             loopi(N->node_tris_num) {
                              if (!(N->node_tris[i]->done)) {
                               N->node_tris[i]->done=true; 
                               col_tris[col_tris_num]=N->node_tris[i];
                               col_tris_num++;
                              }
                             }
                            }
                           }
 }

 // ak sa dosiahne maximalna hlbka, rekurzia pokracuje na predoslej hlbke
 if (tree_level==TREE_DEPTH) { tree_level--; return(0); }

 tree_level++; select_col_tris(N->node[0]);
 tree_level++; select_col_tris(N->node[1]);
 tree_level++; select_col_tris(N->node[2]);
 tree_level++; select_col_tris(N->node[3]);
 tree_level++; select_col_tris(N->node[4]);
 tree_level++; select_col_tris(N->node[5]);
 tree_level++; select_col_tris(N->node[6]);
 tree_level++; select_col_tris(N->node[7]);

 // rekurzia pokracuje na predoslej hlbke
 tree_level--; return(0);
}
