#ifndef _OCTREE_H_
#define _OCTREE_H_

// maximalna hlbka oktanoveho stromu
#define TREE_DEPTH 4

typedef struct __tree_node {            // list oktanoveho stromu
 GLuint id;                             // identifikator listu

 GLfloat minx,maxx,miny,maxy,minz,maxz; // rozmery listu (AABB)

 _tri **node_tris;                      // pole odkazov na trojuholniky priradene listu
 GLuint node_tris_num;                  // pocet trojuholnikov priradenych listu

 __tree_node *node[8];                  // odkazy na 8 listov (podlistov) aktualneho listu

 bool view_bit;                         // indikator viditelnosti listu
} _tree_node;

// trieda premennych a funkcii (objektov) pre pracu s oktanovym stromom
class COctree {
public:
 COctree();          // konstruktor
 virtual ~COctree(); // destruktor

 GLuint nodes_num; // pocet listov oktanoveho stromu
 GLuint tris_num;  // pocet odkazov na trojuholniky, ktore su ulozene v oktanovom strome
 
 GLuint view_tris_num; // pocet aktivnych ukazovatelov na viditelne trojuholniky
 GLuint tests_num;     // pocet testov na viditelnost listov oktanoveho stromu
 GLuint col_tris_num;  // pocet aktivnych ukazovatelov na kolizne trojuholniky

 // vytvorenie oktanoveho stromu
 int create(void);
 // vykreslenie oktanoveho stromu
 void draw(void);
 // zrusenie oktanoveho stromu
 int destroy(void);
 // vyber viditelnych a koliznych trojuholnikov
 void select_tris(vec3 AP,GLfloat AE,GLfloat AW,bool S);
private:
 // pocitadla vyhradenej/uvolnenej pamate
 GLuint alloc_mem,free_mem;

 _tree_node *otree; // rodicovsky list oktanoveho stromu

 int a; // pomocne premenne
 vec3 ap;
 GLfloat ae,aw;

 GLuint tree_level; // aktualna hlbka oktanoveho stromu
 GLuint del_level;  // hlbka oktanoveho stromu, v ktorej budu zmazane listy

 // pridanie listov do oktanoveho stromu - REKURZIVNA FUNKCIA
 int node_add(_tree_node *N);
 // zaradenie trojuholnikov do listov oktanoveho stromu - REKURZIVNA FUNKCIA
 int tris_add(_tree_node *N);
 // vykreslenie listov oktanoveho stromu - REKURZIVNA FUNKCIA
 int node_draw(_tree_node *N);
 // zrusenie listov oktanoveho stromu - REKURZIVNA FUNKCIA
 int node_del(_tree_node *N);

 // vyber viditelnych trojuholnikov - REKURZIVNA FUNKCIA
 int select_view_tris(_tree_node *N);
 // vyber koliznych trojuholnikov - REKURZIVNA FUNKCIA
 int select_col_tris(_tree_node *N);
};

#endif
