#ifndef _PLANTTREE_H_
#define _PLANTTREE_H_

// pocet billboard textur
#define SPRITES_NUM 8

// hranicna vzdialenost pre vykreslenie 3D stromu
#define THRESHOLD1 20.0f
// hranicna vzdialenost pre vykreslenie 2D billboardu
#define THRESHOLD2 30.0f
// hranicna vzdialenost pre vykreslenie stromu
#define THRESHOLD3 40.0f

// kostra stromu
struct skeleton
{
 // zaciatok a koniec ciary
 vec3 s,e;
 // hlbka rekurzie
 int d;
 // dlzka ciary
 float l;
 // indikator ci ma dana cast kostry stromu priameho "rodica"
 bool is_parent;
 // pocet segmentov ciary
 int segs;
 // prirastky zmien poloh zaciatkov a koncov jednotlivych segmentov
 // je mozne vytvorit max. 3 segmenty!!!
 vec3 sgs[3];
 // maximalne 8 vrcholov, z ktorych sa vytvori horny koniec
 // danej casti modelu stromu
 vec3 topP[8];
};

// konar stromu
struct branch
{
 // konstruktor
 branch() { left=middle=right=NULL; }
 // destruktor
 ~branch() { }

 // startovacia poloha konara, dlzka, uhol natocenia okolo osi Z
 float start,length,angle;
 // lavy, stredny a pravy "potomok" konara
 branch *left,*middle,*right;
 // matica obsahujuca polohu a natocenie konara
 mat16 m;
};

// list stromu
struct leaf
{
 vec3 pos,dir; // poloha, natocenie
 vec3 bend;    // ohnutie listu
};

// trieda premennych a funkcii pre pracu s 3D stromom
class CPlanttree {
public:
 // konstruktor
 CPlanttree() {
  // kmen stromu - 1. cast kostry stromu
  stem=NULL;
   // reset prirastkov zmien poloh jednotlivych segmentov
  Sline.sgs[0].set(0.0f,0.0f,0.0f);
  Sline.sgs[1].set(0.0f,0.0f,0.0f);
  Sline.sgs[2].set(0.0f,0.0f,0.0f);
  // inicializacia generatora nahodnych cisiel
  srand(0);
 };
 // destructor
 ~CPlanttree() {};

 // vytvorenie stromu
 int generate(int MAXDEPTH,
              float LENGTH,float LENGTHMOD,
              float ANGLE,float ANGLEMOD,
              float BRANCHWIDTH,float LEAFSIZE,
              GLuint BRANCHTEX,GLuint LEAFTEX);
 // vykreslenie kostry stromu
 void drawskeleton(void);
 // vykreslenie stromu
 void draw(float PX,float PY,float PZ);
 // ziskanie trojuholnikov patriacich stromu
 int get_tris(vec3 TP);
private:
 // identifikatory textur pre konare (aj kmen), listy a billboard stromu
 GLuint branch_texture,leaf_texture,sprite_texture[SPRITES_NUM];
 // hlbka rekurzie
 int Tmaxdepth;
 // dlzka casti kostry stromu, koeficient kratenia dlzky
 float Tlength,Tlengthmod;
 // uhol natocenia casti kostry stromu, koeficient zmeny natocenia
 float Tangle,Tanglemod;
 // hrubka konara, velkost listu
 float Tbranchwidth,Tleafsize;
 // pole ciar - casti kostry stromu a pole listov stromu
 vector<skeleton> lines;
 vector<leaf> leaves;
 // kmen stromu - 1. cast kostry stromu
 branch *stem;
 // pomocne premenne
 int i,j,k,l,mem;
 float r,lim,inc,r1,inc1;
 // pomocne vektory
 vec3 Vpos,Vrot,Vdir,Vdirnorm,Vup,Vright,Vstart,Vend,Vbend;
 vec3 PA,PB,PC,PD;
 float Vsize;
 // pomocna matica
 float mat[16];
 // matica 4x4
 mat16 mv_matrix;
 // pomocne vektory
 vec3 VA,VB;
 // ciara patriaca kostre stromu
 skeleton Sline,L;
 // listy stromu
 leaf lf1,lf2;
 // velkost spodneho a vrchneho konca konara
 float botrad,toprad;
 // indikacia existencie "rodica" casti kostry stromu
 bool parent;
 // vzdialenost od stromu
 float dist;

 vec3 TREEPOS;

 // vytvorenie jednotlivych casti kostry stromu
 // REKURZIVNA FUNKCIA!!!
 int create_skeleton(branch *B,mat16 M,
                     int DEPTH,
                     float LENGTH,float ANGLE);
 // uvolnenie vyhradeneho pamatoveho miesta
 // REKURZIVNA FUNKCIA!!!
 int delete_skeleton(branch *B,int DEPTH);
 // vykreslenie 3D stromu
 void RenderTree();
 // vykreslenie konara
 void RenderBranch(unsigned int I,float BOTRAD,float TOPRAD,float SEGS);
 // priprava 2D billboard textur
 void RenderSprite(int S);
 
 // ziskanie trojuholnikov konara
 void get_branch(unsigned int I,float BOTRAD,float TOPRAD,float SEGS);
};

#endif
