// hlavickove subory *******************************************************
#include "../includes.h"

CShader fl_vp; // flashlight VertexProgram
CShader fl_fp; // flashlight FragmentProgram

CShader ter_vp; // terrain VertexProgram
CShader ter_fp; // terrain FragmentProgram

CShader nm_fl_vp; // normalmap & flashlight VertexProgram
CShader nm_fl_fp; // normalmap & flashlight FragmentProgram

CShader decal_vp; // decal VertexProgram
CShader decal_fp; // decal FragmentProgram

CShader glsl_shader; // GLSL vertex & fragment shader

CShader lighting_shader; // GLSL lighting shader

// konstruktor *************************************************************
CShader::CShader()
{
 sh_code=NULL; target=0; id=0; err_line=-1;

 loopi(GLSL_OBJ_NUM) glsl_object[i]=0;
 loc=-1;
}

// destruktor **************************************************************
CShader::~CShader()
{
 // zrusenie objektu ShaderProgramu
 glDeletePrograms_(1,&id);
 // zrusenie program objektov GLSL shadera
 loopi(GLSL_OBJ_NUM) {
  if (glsl_object[i]!=0) { glDeleteObject_(glsl_object[i]); glsl_object[i]=0; }
 }
}

// nacitanie kodu ShaderProgramu *******************************************
int CShader::PRGload(const char *SRC)
{
 size_t result;
 // otvorenie suboru so ShaderProgramom
 FILE *file=fopen(SRC,"r");
 // nacitanie obsahu suboru do premennej sh_code
 if (file) {
  // velkost suboru
  fseek(file,0,SEEK_END); int size=ftell(file);
  // vyhradenie pamate pre ulozenie obsahu suboru
  sh_code=new char[size+1];
  if (sh_code==NULL) {
   // zatvorenie suboru so ShaderProgramom
   fclose(file);
   target=0; id=0; err_line=-1;
   printf("Nemozno vyhradit pamat pre ulozenie obsahu suboru!\n"); return(1);
  }
  // nacitanie obsahu suboru
  fseek(file,0,SEEK_SET); result=fread(sh_code,1,size,file); sh_code[size]='\0';
  if (result==0) {
   DELETEA(sh_code); target=0; id=0; err_line=-1;
   printf("Chyba nacitania obsahu suboru %s!\n",SRC); return(1);
  }
  // zatvorenie suboru so ShaderProgramom
  fclose(file);
 } else {
    DELETEA(sh_code); target=0; id=0; err_line=-1;
    printf("Chyba nacitania obsahu suboru %s!\n",SRC); return(1);
   }

 err_line=-1;

 // vytvorenie objektu ShaderProgramu
 if (!strncmp(sh_code,"!!ARBvp1.0",10)) { // ARB VertexProgram
  target=GL_VERTEX_PROGRAM_ARB;
 } else if (!strncmp(sh_code,"!!ARBfp1.0",10)) { // ARB FragmentProgram
         target=GL_FRAGMENT_PROGRAM_ARB;
        } else { // nepodporovany ShaderProgram
                DELETEA(sh_code); target=0; id=0; err_line=-1;
                printf("Nepodporovany ShaderProgram!\n"); return(1);
               }

 // vytvorenie objektu ShaderProgramu
 glGenPrograms_(1,&id);
 glBindProgram_(target,id);
 // umiestnenie nacitaneho kodu do objektu ShaderProgramu
 glProgramString_(target,GL_PROGRAM_FORMAT_ASCII_ARB,strlen(sh_code),sh_code);
 // detekcia chyby
 glGetIntegerv(GL_PROGRAM_ERROR_POSITION_ARB,&err_line);

 // ak bola detekovana chyba, premenna err_line obsahuje cislo chybneho riadku ShaderProgramu
 if (err_line!=-1) {
  int line=0;      // cislo riadku
  char *s=sh_code; // pomocna premenna
  while (err_line-- && *s) { if (*s++ == '\n') line++; }
  while (s >= sh_code && *s != '\n') s--;
  DELETEA(sh_code); target=0; id=0; err_line=-1;
  printf("Chyba na %i. riadku ShaderProgramu: %s!\n",line,s); return(1);
 }

 printf("%s ",SRC);
 // uvolnenie vyhradenej pamate
 DELETEA(sh_code);

 return(0);
}

// zapnutie ShaderProgramu *************************************************
void CShader::PRGenable(void)
{
 glEnable(target);
}

// vypnutie ShaderProgramu *************************************************
void CShader::PRGdisable(void)
{
 glDisable(target);
}

// pripojenie ShaderProgramu ***********************************************
void CShader::PRGbind(void)
{
 glBindProgram_(target,id);
}

// vlozenie Environment parametra do ARB ShaderProgramu ********************
void CShader::PRGenv_param(const GLuint I,const vec4 P)
{
 v[0]=P.x; v[1]=P.y; v[2]=P.z; v[3]=P.w;
 glProgramEnvParameter4fv_(target,I,v);
}

// vlozenie Local parametra do ARB ShaderProgramu **************************
void CShader::PRGlocal_param(const GLuint I,const vec4 P)
{
 v[0]=P.x; v[1]=P.y; v[2]=P.z; v[3]=P.w;
 glProgramLocalParameter4fv_(target,I,v);
}

// nacitanie obsahu textoveho suboru ***************************************
char *CShader::TextFileRead(const char *FNAME)
{
 FILE *f_p=NULL;
 char *f_content=NULL;
 int f_size=0;

 // ziskanie velkosti suboru
 f_p=fopen(FNAME,"rb");
 if (f_p==NULL) { printf("Nemozno otvorit subor: %s!\n",FNAME); return(NULL); }
 fseek(f_p,0,SEEK_END);
 f_size=ftell(f_p);
 fclose(f_p);

 // ak subor nie je prazdny
 if (f_size>0) {
  f_p=fopen(FNAME,"rt");
  // nacitanie obsahu textoveho suboru
  f_content=new char[f_size+1];
  if (f_content==NULL) { printf("Nemozno vyhradit pamat!\n"); fclose(f_p); return(NULL); }
  f_size=(int) fread(f_content,sizeof(char),f_size,f_p);
  f_content[f_size]='\0';
  fclose(f_p);
 }
 else { printf("Subor je prazdny!\n"); return(NULL); }

 return(f_content);
}

// nacitanie kodu GLSL vertex & fragment shadera ***************************
int CShader::GLSLload(const int ID,const char *VP,const char *FP)
{
 GLhandleARB vertex_shader,fragment_shader;
 GLint t=0;

 char *vp_src=TextFileRead(VP); // kod vertex programu
 char *fp_src=TextFileRead(FP); // kod fragment programu

 if (vp_src==NULL || fp_src==NULL) {
  printf("Nemozno nacitat GLSL shader!\n");
  DELETEA(vp_src); DELETEA(fp_src);
  return(1);
 }

 // vytvorenie program objektu GLSL shadera
 glsl_object[ID]=glCreateProgramObject_();

 // vytvorenie objektov vertex & fragment shadera
 vertex_shader=glCreateShaderObject_(GL_VERTEX_SHADER_ARB);
 fragment_shader=glCreateShaderObject_(GL_FRAGMENT_SHADER_ARB);

 // nastavenie kodu vertex & fragment shadera
 glShaderSource_(vertex_shader,1,(const char**) &vp_src,NULL);
 glShaderSource_(fragment_shader,1,(const char**) &fp_src,NULL);

 // kompilacia shader programov
 glCompileShader_(vertex_shader);
 glGetObjectParameteriv_(vertex_shader,GL_OBJECT_COMPILE_STATUS_ARB,&t);
 if (t==0) {
  DELETEA(vp_src); DELETEA(fp_src);
  glDeleteObject_(glsl_object[ID]); glsl_object[ID]=0; return(1);
 }
 glCompileShader_(fragment_shader);
 glGetObjectParameteriv_(fragment_shader,GL_OBJECT_COMPILE_STATUS_ARB,&t);
 if (t==0) {
  DELETEA(vp_src); DELETEA(fp_src);
  glDeleteObject_(glsl_object[ID]); glsl_object[ID]=0; return(1);
 }

 // linkovanie shader programov
 glAttachObject_(glsl_object[ID],vertex_shader);
 glAttachObject_(glsl_object[ID],fragment_shader);
 glLinkProgram_(glsl_object[ID]);
 glGetObjectParameteriv_(glsl_object[ID],GL_OBJECT_LINK_STATUS_ARB,&t);
 if (t==0) {
  DELETEA(vp_src); DELETEA(fp_src);
  glDeleteObject_(glsl_object[ID]); glsl_object[ID]=0; return(1);
 }

 printf("%s %s ",VP,FP);
 // uvolnenie pamate vyhradenej pre kod shaderov
 DELETEA(vp_src); DELETEA(fp_src);

 return(0);
}

// zapnutie GLSL shadera ***************************************************
void CShader::GLSLon(const int ID)
{
 glUseProgramObject_(glsl_object[ID]);
}

// vypnutie GLSL shadera ***************************************************
void CShader::GLSLoff(void)
{
 glUseProgramObject_(0);
}

// nastavenie parametrov GLSL shadera **************************************
void CShader::GLSLuniformI(const int ID,const char *NAME,const int I)
{
 loc=glGetUniformLocation_(glsl_object[ID],NAME);
 if (loc!=-1) glUniform1i_(loc,I);
}

void CShader::GLSLuniformF(const int ID,const char *NAME,const float F0,const float F1,const float F2,const float F3)
{
 loc=glGetUniformLocation_(glsl_object[ID],NAME);
 if (loc!=-1) glUniform4f_(loc,F0,F1,F2,F3);
}

void CShader::GLSLuniformmatrixFV(const int ID,const char *NAME,const float *FV)
{
 loc=glGetUniformLocation_(glsl_object[ID],NAME);
 if (loc!=-1) glUniformMatrix4fv_(loc,1,GL_FALSE,FV);
}
