// hlavickove subory *******************************************************
#include "../includes.h"
// vlastny .h subor
#include "skydome.h"

// konstruktor *************************************************************
CSkydome::CSkydome()
{
 // polomer, splostenie Skydomu
 SkyRadius=450.0f; Dampening=0.8f;
 // pocet pruhov, pocet stien, pocet indikatorov
 NumSlices=8; NumSides=16; NumIndices=NumSlices*(NumSides+1)*2;

 // velkost slnka, mesiaca
 SunSize=60.0f; MoonSize=40.0f;

 vertices=NULL; VB=NULL;
 colors=NULL; CB=NULL;
 indexes=NULL; IB=NULL;

 // pocet hviezd
 NumStars=1000;

 stars=NULL;
 starscolor=NULL;

 // pociatocne nastavenie casu pre pohyb oblakov
 clouds_time=0.0f;
 // objekt zakrivenej plochy
 CPVertices=NULL; CPVerticesNum=0;
 CPIndices=NULL; CPIndicesNum=0;
 // nastavenie poctu poli objektu zakrivenej plochy
 CPsides=8;
 // nastavenie poctu opakovania sa textury objektu oblakov
 CPuv=4.0f;
}

// destruktor **************************************************************
CSkydome::~CSkydome()
{
 // zrusenie oblohy
 destroy();
}

// inicializacia oblohy ****************************************************
int CSkydome::init(float Lat,float Long,float Time,float Day)
{
 Latitude=DEG2RAD(Lat);    // zemepisna vyska
 Longitude=DEG2RAD(Long);  // zemepisna sirka
 TimeOfDay=(Time)*3600.0f; // hodina
 JulianDay=Day;            // den

 NumVertices=(NumSlices+1)*(NumSides+1);
 // pole vrcholov
 vertices=new vec3[NumVertices];
 VB=new float[3*NumVertices];
 // pole farieb
 colors=new col4[NumVertices];
 CB=new float[4*NumVertices];
 // pole indexov
 indexes=new unsigned int[NumIndices];
 IB=new unsigned int[NumIndices];

 // uhol medzi dvoma horizontalnymi vrcholmi Skydomu
 // predstavuje sirku steny Skydomu
 polyAng=2.0f*PI/NumSides;

 k=0;
 // vytvorenie Skydomu
 for (j=0;j<=NumSlices;j++) { // pruhy
  ang=(float) j*((PI/2.0f)/(float) NumSlices);
  for (i=0;i<=NumSides;i++) { // steny
   vx=cos((float) i*polyAng)*cos(ang); // vrchol X
   vy=sin(ang)*Dampening;              // vrchol Y
   vz=sin((float) i*polyAng)*cos(ang); // vrchol Z
   // ulozenie vrcholu do pola vrcholov
   vertices[i+j*(NumSides+1)].x=vx*SkyRadius;
   vertices[i+j*(NumSides+1)].y=vy*SkyRadius;
   vertices[i+j*(NumSides+1)].z=vz*SkyRadius;
   // priprava vrcholov pre VBO
   VB[k++]=vx*SkyRadius;
   VB[k++]=vy*SkyRadius;
   VB[k++]=vz*SkyRadius;
  }
 }

 // INFO:
 //fprintf(stderr,"NumVertices=%u k=%u NumIndices=%u\n",NumVertices,k,NumIndices);

 k=0;
 // priprava indexov
 ind=0;
 for (j=1;j<=NumSlices;j++) { // pruhy
  for (i=0;i<=NumSides;i++) { // steny
   indexes[ind++]=i+j*(NumSides+1);
   indexes[ind++]=i+(j-1)*(NumSides+1);
   // priprava indexov pre VBO
   IB[k++]=i+j*(NumSides+1);
   IB[k++]=i+(j-1)*(NumSides+1);
  }
 }

 // pre nacitanie obrazkov
 SDL_Surface *TextureImage;

 // nacitanie textury slnka
 TextureImage=IMG_Load("data/textures/sky/sun.tga");
 if (TextureImage==NULL) {
  fprintf(stderr,"Chyba nacitania textury slnka!\n"); return(1);
 }
 glGenTextures(1,&SunTex);
 glBindTexture(GL_TEXTURE_2D,SunTex);
 glTexImage2D(GL_TEXTURE_2D,0,GL_RGB,TextureImage->w,TextureImage->h,0,GL_BGR,
              GL_UNSIGNED_BYTE,TextureImage->pixels);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);

 if (TextureImage) SDL_FreeSurface(TextureImage); TextureImage=NULL;

 // nacitanie textury mesiaca
 TextureImage=IMG_Load("data/textures/sky/moon.tga");
 if (TextureImage==NULL) {
  fprintf(stderr,"Chyba nacitania textury mesiaca!\n"); return(1);
 }
 glGenTextures(1,&MoonTex);
 glBindTexture(GL_TEXTURE_2D,MoonTex);
 glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA,TextureImage->w,TextureImage->h,0,GL_BGRA,
              GL_UNSIGNED_BYTE,TextureImage->pixels);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);

 if (TextureImage) SDL_FreeSurface(TextureImage); TextureImage=NULL;

 // nacitanie textury oblakov
 TextureImage=IMG_Load("data/textures/sky/clouds.tga");
 if (TextureImage==NULL) {
  fprintf(stderr,"Chyba nacitania textury oblakov!\n"); return(1);
 }
 glGenTextures(1,&CloudsTex);
 glBindTexture(GL_TEXTURE_2D,CloudsTex);
 glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA,TextureImage->w,TextureImage->h,0,GL_BGRA,
              GL_UNSIGNED_BYTE,TextureImage->pixels);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);

 if (TextureImage) SDL_FreeSurface(TextureImage); TextureImage=NULL;

 // priprava hviezd
 stars=new vec3[NumStars];         // pole vrcholov - hviezd
 starscolor=new col4[NumStars]; // pole farieb - hviezd
 for (i=0;i<NumStars;i++) {
  st=(rand()%90)/180.0f*PI;  // uhol theta
  sp=(rand()%360)/180.0f*PI; // uhol phi

  stars[i].x=cos(sp)*cos(st)*SkyRadius;   // vrchol X
  stars[i].y=sin(st)*SkyRadius*Dampening; // vrchol Y
  stars[i].z=sin(sp)*cos(st)*SkyRadius;   // vrchol Z

  c=(rand()%256)/255.0f; // jas hviezdy
  starscolor[i]=col4(c,c,c,c);
 }

 // DisplayList pre objekt oblakov
 clouds_DL=0;
 // vytvorenie objektu oblakov
 clouds_create(500.0f,150.0f,0.99f);

 // priprava VBO objektu
 glGenBuffers(1,&VBObuffer);   // pole vrcholov a farieb
 glGenBuffers(1,&VBOi_buffer); // pole indexov

 update(0); // aktualizacia oblohy

 return(0);
}

// zrusenie oblohy *********************************************************
void CSkydome::destroy(void)
{
 // zrusenie objektov textur
 glDeleteTextures(1,&SunTex);
 glDeleteTextures(1,&MoonTex);
 glDeleteTextures(1,&CloudsTex);
 
 // zrusenie VBO bufferov
 glDeleteBuffers(1,&VBObuffer);
 glDeleteBuffers(1,&VBOi_buffer);

 // zrusenie DisplayListu pre objekt oblakov
 glDeleteLists(clouds_DL,1);
 clouds_DL=0;
 // zrusenie objektu oblakov
 clouds_destroy();

 // hviezdy
 DELETEA(starscolor);
 DELETEA(stars);

 // Skydome
 DELETEA(IB);
 DELETEA(indexes);
 DELETEA(CB);
 DELETEA(colors);
 DELETEA(VB);
 DELETEA(vertices);
}

// aktualizacia oblohy *****************************************************
void CSkydome::update(float deltatime)
{
 TimeOfDay+=deltatime;          // nova hodina
 if (TimeOfDay>3600.0f*24.0f) { // novy den
  JulianDay+=1.0f;
  TimeOfDay-=3600.0f*24.0f;
 }

 hour=TimeOfDay/3600.0f;
 solarTime=hour+0.170f*sin(4.0f*PI*(JulianDay-80.0f)/373.0f)-0.129f*
                       sin(2.0f*PI*(JulianDay-8.0f)/355.0f)+12.0f*
                       (Latitude-Longitude)/PI;
 declination=0.4093f*sin(2.0f*PI*(JulianDay-81.0f)/368.0f);

 // mesiac
 MoonTheta=PI/3.0f;
 MoonPhi=PI/2.0f;
 // slnko
 SunTheta=PI/2.0f-asin(sin(Latitude)*sin(declination)-cos(Latitude)*cos(declination)*cos(PI*solarTime/12.0f));
 SunPhi=atan(-cos(declination)*sin(PI*solarTime/12.0f)/
             (cos(Latitude)*cos(declination)-sin(Latitude)*sin(declination)*sin(PI*solarTime/12.0f)));

 if (SunPhi<0.0f) SunPhi=2.0f*PI+SunPhi;

 SliceAng=(PI/2.0f)/NumSlices;
 SideAng=(PI*2.0f)/NumSides;

 SkyShader.SetInfo(SunTheta,2.0f);

 ind=0; k=0;
 for (j=0;j<=NumSlices;j++) {
  for (i=0;i<=NumSides;i++) {
   theta=PI/2.0f-SliceAng*j;
   phi=i*SideAng;

   SkyShader.GetVertexColor(theta,phi,SunPhi,&colors[ind]);
   // priprava farieb pre VBO
   CB[k++]=colors[ind].R;
   CB[k++]=colors[ind].G;
   CB[k++]=colors[ind].B;
   CB[k++]=colors[ind].A;
   ind++;
  }
 }

 glBindBuffer(GL_ARRAY_BUFFER,VBObuffer);
 glBufferData(GL_ARRAY_BUFFER,3*NumVertices*sizeof(float)+
                              4*NumVertices*sizeof(float),NULL,GL_STATIC_DRAW);
 glBufferSubData(GL_ARRAY_BUFFER,0,3*NumVertices*sizeof(float),VB);
 glBufferSubData(GL_ARRAY_BUFFER,3*NumVertices*sizeof(float),
                                 4*NumVertices*sizeof(float),CB);

 glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,VBOi_buffer);
 glBufferData(GL_ELEMENT_ARRAY_BUFFER,NumIndices*sizeof(unsigned int),
                                      IB,GL_STATIC_DRAW);

 // ziskanie polohy mesiaca
 MoonPos.set(sin(MoonTheta)*cos(MoonPhi)*SkyRadius,
             cos(MoonTheta)*SkyRadius,
             sin(MoonTheta)*sin(MoonPhi)*SkyRadius);

 // ziskanie polohy slnka
 SunPos.set(sin(SunTheta)*cos(SunPhi)*SkyRadius,
            cos(SunTheta)*SkyRadius,
            sin(SunTheta)*sin(SunPhi)*SkyRadius);
}

// vykreslenie oblohy ******************************************************
void CSkydome::draw(GLfloat DT)
{
 // ulozenie potrebnych atributov
 glPushAttrib(GL_DEPTH_BUFFER_BIT|GL_TEXTURE_BIT|GL_LIGHTING_BIT);
 glDisable(GL_LIGHTING);
 glDisable(GL_TEXTURE_2D);

 glPushMatrix();

 // posun Skydomu vzhladom k vyske hladiny mora
 glTranslatef(0.0f,-25.0f,0.0f);

 ////glDisable(GL_DEPTH_TEST);
 ////RenderElements(true);  // vykreslenie mesiaca a hviezd

 ////glEnable(GL_DEPTH_TEST);
 RenderSkydome();       // vykreslenie oblohy

 glDisable(GL_DEPTH_TEST);
 RenderElements(false); // vykreslenie slnka

 clouds_draw(DT);       // vykreslenie oblakov

 glPopMatrix();

 glPopAttrib(); // vyber ulozenych atributov
 
 glBindBuffer(GL_ARRAY_BUFFER,0);
}

// vykreslenie mesiaca a hviezd / slnka ************************************
void CSkydome::RenderElements(bool moon)
{
 glGetFloatv(GL_MODELVIEW_MATRIX,mat); // pre kreslenie technikou billboard

 // ziskanie vektorov rovnobeznych s osami X a Y pohladu
 vertx.set(mat[0],mat[4],mat[8]);
 verty.set(mat[1],mat[5],mat[9]);

 if (moon) { // vykreslenie mesiaca a hviezd
  glDisable(GL_TEXTURE_2D);
  glEnableClientState(GL_VERTEX_ARRAY);
  glEnableClientState(GL_COLOR_ARRAY);

  glVertexPointer(3,GL_FLOAT,0,stars);
  glColorPointer(4,GL_FLOAT,0,starscolor);
  glDrawArrays(GL_POINTS,0,NumStars); // vykreslenie hviezd

  glDisableClientState(GL_VERTEX_ARRAY);
  glDisableClientState(GL_COLOR_ARRAY);

  glColor4f(1.0f,1.0f,1.0f,1.0f); // standardna farba - biela

  Vertices[0]=(MoonPos+(-vertx-verty)*MoonSize);
  Vertices[1]=(MoonPos+( vertx-verty)*MoonSize);
  Vertices[2]=(MoonPos+( vertx+verty)*MoonSize);
  Vertices[3]=(MoonPos+(-vertx+verty)*MoonSize);

  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D,MoonTex);

  glEnable(GL_ALPHA_TEST);
  glAlphaFunc(GL_GREATER,0.0f);
  glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);

  // vykreslenie mesiaca
  glBegin(GL_QUADS);
   glTexCoord2f(0.0f,0.0f); glVertex3fv(Vertices[0].v);
   glTexCoord2f(1.0f,0.0f); glVertex3fv(Vertices[1].v);
   glTexCoord2f(1.0f,1.0f); glVertex3fv(Vertices[2].v);
   glTexCoord2f(0.0f,1.0f); glVertex3fv(Vertices[3].v);
  glEnd();

  glDisable(GL_BLEND);

  glDisable(GL_TEXTURE_2D);
 }
 else { // vykreslenie slnka
  glColor4f(1.0f,1.0f,1.0f,0.8f); // polopriesvitna standardna farba - biela

  Vertices[0]=(SunPos+(-vertx-verty)*SunSize);
  Vertices[1]=(SunPos+( vertx-verty)*SunSize);
  Vertices[2]=(SunPos+( vertx+verty)*SunSize);
  Vertices[3]=(SunPos+(-vertx+verty)*SunSize);

  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D,SunTex);

  glEnable(GL_BLEND);
  glBlendFunc(GL_ONE,GL_ONE);

  // vykreslenie slnka
  glBegin(GL_QUADS);
   glTexCoord2f(0.0f,0.0f); glVertex3fv(Vertices[0].v);
   glTexCoord2f(1.0f,0.0f); glVertex3fv(Vertices[1].v);
   glTexCoord2f(1.0f,1.0f); glVertex3fv(Vertices[2].v);
   glTexCoord2f(0.0f,1.0f); glVertex3fv(Vertices[3].v);
  glEnd();

  glDisable(GL_BLEND);

  glDisable(GL_TEXTURE_2D);
 }

 glColor4f(1.0f,1.0f,1.0f,1.0f); // standardna farba - biela
}

// vykreslenie oblohy ******************************************************
void CSkydome::RenderSkydome(void)
{
 ////glEnable(GL_BLEND);
 ////glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
/*
 glEnableClientState(GL_VERTEX_ARRAY);
 glVertexPointer(3,GL_FLOAT,0,&vertices[0]);

 glEnableClientState(GL_COLOR_ARRAY);
 glColorPointer(4,GL_FLOAT,0,&colors[0]);

 for (i=0;i<NumSlices;i++) {
  glDrawElements(GL_TRIANGLE_STRIP,(NumSides+1)*2,GL_UNSIGNED_INT,&indexes[i*(NumSides+1)*2]);
 }

 glDisableClientState(GL_VERTEX_ARRAY);
 glDisableClientState(GL_COLOR_ARRAY);
*/
 // vykreslenie VBO objektu
 glEnableClientState(GL_VERTEX_ARRAY);
 glEnableClientState(GL_COLOR_ARRAY);

 glBindBuffer(GL_ARRAY_BUFFER,VBObuffer);
 glVertexPointer(3,GL_FLOAT,0,(void*)0);
 glColorPointer(4,GL_FLOAT,0,(void*)(3*NumVertices*sizeof(float)));
 glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,VBOi_buffer);
 glDrawElements(GL_TRIANGLE_STRIP,NumIndices,GL_UNSIGNED_INT,0);

 glDisableClientState(GL_VERTEX_ARRAY);
 glDisableClientState(GL_COLOR_ARRAY);

 ////glDisable(GL_BLEND);

 glColor4f(1.0f,1.0f,1.0f,1.0f); // standardna farba - biela
}

// vytvorenie objektu oblakov **********************************************
int CSkydome::clouds_create(GLfloat R1,GLfloat R2,GLfloat A)
{
 // vypocet velkosti zakrivenej plochy
 CPsize=(GLfloat) sqrt(((2.0f*R1)*(2.0f*R1))*0.5f);
 // vypocet velkosti 1 pola plochy - prirastku suradnic vrcholov
 Vdelta=CPsize/(GLfloat) CPsides;
 // vypocet prirastku texturovacich suradnic
 Tdelta=1.0f/(GLfloat) CPsides;
 // koeficient pre vypocet vysky vrcholu - v zavislosti od vzdialenosti vrcholu od stredu plochy
 R1sq=R1*R1;
 // koeficient pre vypocet priesvitnosti - v zavislosti od vzdialenosti vrcholu od stredu plochy
 Asq=(CPsize*A*0.5f)*(CPsize*A*0.5f);

 // pocet vrcholov zakrivenej plochy
 CPVerticesNum=(CPsides+1)*(CPsides+1);
 // odkazy na vrcholy - 1 stvoruholnikova plocha - 2 trojuholiky, pre kazdy 3 vrcholy
 CPIndicesNum=CPsides*CPsides*2*3;

 // vyhradenie pamatoveho miesta pre vrcholy zakrivenej plochy
 CPVertices=(clouds_plane *) malloc(CPVerticesNum*sizeof(clouds_plane));
 // vyhradenie pamatoveho miesta pre odkazy na jednotlive vrcholy
 CPIndices=(GLuint *) malloc(CPIndicesNum*sizeof(GLuint));
 if ((CPVertices==NULL)||(CPIndices==NULL)) {
  fprintf(stderr,"Chyba vyhradenia pamate pre objekt oblakov!\n"); return(1);
 }

 // vypocet polohy vrcholov, texturovacich suradnic, priesvitnosti
 for (j=0;j<=CPsides;j++) {
  for (i=0;i<=CPsides;i++) {
   // vypocet suradnic x a z
   CPv.x=(-0.5f*CPsize)+((GLfloat) i*Vdelta);
   CPv.z=(-0.5f*CPsize)+((GLfloat) j*Vdelta);
   // vypocet suradnice y
   CPv.y=(CPv.x*CPv.x+CPv.z*CPv.z)/R1sq;
   CPv.y=(1.0f-CPv.y)*R2;
   // vypocet priesvitnosti textury v danom vrchole
   CPv.a=(CPv.x*CPv.x+CPv.z*CPv.z)/Asq;
   CPv.a=1.0f-CPv.a*CPv.a*CPv.a;
   if (CPv.a<0.0f) CPv.a=0.0f;
   // vypocet texturovacich suradnic
   CPv.u=CPuv*((GLfloat) i*Tdelta);
   CPv.v=CPuv*((GLfloat) j*Tdelta);
   // ulozenie vypocitanych hodnot do pola
   CPVertices[i+j*(CPsides+1)]=CPv;
  }
 }

 // vytvorenie pola odkazov na konkretny vrchol
 // 2 susedne trojuholiky = 1 stvoruholnik
 for (j=0;j<CPsides;j++) {
  for (i=0;i<CPsides;i++) {
   b=(i+j*(CPsides+1));
   // 1. trojuholik
   CPIndices[a++]=b; CPIndices[a++]=b+1; CPIndices[a++]=b+CPsides+1;
   // 2. trojuholik
   CPIndices[a++]=b+1; CPIndices[a++]=b+CPsides+2; CPIndices[a++]=b+CPsides+1;
  }
 }

 return(0);
}

// vykreslenie objektu oblakov *********************************************
void CSkydome::clouds_draw(GLfloat DT)
{
 /*if (clouds_DL==0) {

  clouds_DL=glGenLists(1);
  glNewList(clouds_DL,GL_COMPILE_AND_EXECUTE);*/

  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D,CloudsTex);

  // modifikacia matice textury - pohyb oblakov
  glMatrixMode(GL_TEXTURE);
  glLoadIdentity();
  clouds_time+=DT/50.0f;
  glTranslatef(clouds_time,0.0f,0.0f);

  // textura oblakov je polopriesvitna
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA,GL_ONE);
  glAlphaFunc(GL_GREATER,0.0f);

  // zakrivena plocha je zlozena zo stvoruholnikov, kazdy stvoruholnik
  // je zlozeny z dvoch trojuholnikov
  glBegin(GL_TRIANGLES);
   for (i=0;i<CPIndicesNum;i++) {
    // farba - je dolezita pre urcenie priesvitnosti - alfa hodnota
    glColor4f(1.0f,1.0f,1.0f,CPVertices[CPIndices[i]].a);
    // texturovacie suradnice
    glTexCoord2f(CPVertices[CPIndices[i]].u,
                 CPVertices[CPIndices[i]].v);
    // vrcholy trojuholnikov
    glVertex3f(CPVertices[CPIndices[i]].x,
               CPVertices[CPIndices[i]].y,
               CPVertices[CPIndices[i]].z);
   }
  glEnd();

  glDisable(GL_BLEND); // vypnutie priesvitnosti

  glLoadIdentity(); // pre texturovu maticu

  glMatrixMode(GL_MODELVIEW); // nastavenie povodnej MODELVIEW matice

  glColor4f(1.0f,1.0f,1.0f,1.0f); // standardna farba - biela

 /* glEndList();
 }
 else glCallList(clouds_DL);*/
}

// zrusenie objektu oblakov ************************************************
void CSkydome::clouds_destroy(void)
{
 if (CPVertices!=NULL) { free(CPVertices); CPVertices=NULL; }
 if (CPIndices!=NULL) { free(CPIndices); CPIndices=NULL; }
}
