#ifndef _SKYDOME_H_
#define _SKYDOME_H_

struct clouds_plane { // zakrivena plocha - objekt oblakov
 GLfloat x,y,z; // 3D vrchol
 GLfloat a;     // priesvitnost
 GLfloat u,v;   // texturovacie suradnice
};

// 4-prvkovy vektor reprezentujuci farbu
class col4 {
public:
 // konstruktory
 col4() { R=G=B=A=0.0f; }
 col4(float CR,float CG,float CB,float CA) {
  if (CR>1.0f) CR=1.0f; if (CG>1.0f) CG=1.0f;  if (CB>1.0f) CB=1.0f; if (CA>1.0f) CA=1.0f;
  if (CR<0.0f) CR=0.0f; if (CG<0.0f) CG=0.0f;  if (CB<0.0f) CB=0.0f; if (CA<0.0f) CA=0.0f;
  R=CR; G=CG; B=CB; A=CA;
 }

 // farba je zlozena z R G B A komponentov
 float R,G,B,A;

 // negacia
 float* operator!() {
  return (float*) this;
 }
 // pricitanie farby (vektora)
 col4 operator+(col4 C) {
  col4 res;
  res.R=R+C.R; res.G=G+C.G; res.B=B+C.B; res.A=A+C.A;
  return(res);
 }
 // odcitanie farby (vektora)
 col4 operator-(col4 C) {
  col4 res;
  res.R=R-C.R; res.G=G-C.G; res.B=B-C.B; res.A=A-C.A;
  return(res);
 }
 // nasobenie skalarom
 col4 operator*(float C) {
  col4 res;
  res.R=R*C; res.G=G*C; res.B=B*C; res.A=A*C;
  return(res);
 }
 // orezanie do <0.0;1.0>
 void clamp(void) {
  if (R>1.0f) R=1.0f; if (G>1.0f) G=1.0f; if (B>1.0f) B=1.0f; if (A>1.0f) A=1.0f;
  if (R<0.0f) R=0.0f; if (G<0.0f) G=0.0f; if (B<0.0f) B=0.0f; if (A<0.0f) A=0.0f;
 }
};

// trieda pre pripravu farby oblohy
class mSkycolor {
public:
 mSkycolor() { T=0.0f; suntheta=0.0f; } // konstruktor
 virtual ~mSkycolor() {}                // destruktor

 // priprava koeficientov pre Perez funkcie
 void SetInfo(float SunTheta,float Turbidity) {
  suntheta=SunTheta; T=Turbidity;
  suntheta2=suntheta*suntheta;
  suntheta3=suntheta*suntheta2;
  T2=T*T;

  Ax=-0.01925f*T-0.25922f; Bx=-0.06651f*T+0.00081f;
  Cx=-0.00041f*T+0.21247f; Dx=-0.06409f*T-0.89887f;
  Ex=-0.00325f*T+0.04517f;

  Ay=-0.01669f*T-0.26078f; By=-0.09495f*T+0.00921f;
  Cy=-0.00792f*T+0.21023f; Dy=-0.04405f*T-1.65369f;
  Ey=-0.01092f*T+0.05291f;

  AY= 0.17872f*T-1.46303f; BY=-0.35540f*T+0.42749f;
  CY=-0.02266f*T+5.32505f; DY= 0.12064f*T-2.57705f;
  EY=-0.06696f*T+0.37027f;

  xz=( 0.00165f*suntheta3-0.00375f*suntheta2+0.00209f*suntheta+0.00000f)*T2+
     (-0.02903f*suntheta3+0.06377f*suntheta2-0.03202f*suntheta+0.00394f)*T+
     ( 0.11693f*suntheta3-0.21196f*suntheta2+0.06052f*suntheta+0.25886f);
  yz=( 0.00275f*suntheta3-0.00610f*suntheta2+0.00317f*suntheta+0.00000f)*T2+
     (-0.04214f*suntheta3+0.08970f*suntheta2-0.04153f*suntheta+0.00516f)*T+
     ( 0.15346f*suntheta3-0.26756f*suntheta2+0.06670f*suntheta+0.26688f);

  X=(4.0f/9.0f-T/120.0f)*(PI-2.0f*suntheta);
  Yz=((4.0453f*T-4.9710f)*tan(X)-0.2155f*T+2.4192f)*1000.0f;
 }

 // ziskanie farby vrcholu
 void GetVertexColor(float th,float phi,float sunphi,col4* Color) {
  gm=Angle(th,phi,suntheta,sunphi);
  if (fabs(th-(PI/2.0f))<0.001f) cosTheta=0.00001f;
  else cosTheta=cos(th);
  cosGamma=cos(gm);
  cosSTheta=cos(suntheta); // uhol medzi horiznotom a polohou slnka

  x=xz*PerezFunction_x(cosTheta,cosGamma,gm);
  y=yz*PerezFunction_y(cosTheta,cosGamma,gm);
  Y=Yz*PerezFunction_Y(cosTheta,cosGamma,gm);

  ToRGBA(x,y,Y,Color);

  // priesvitnost Skydomu v pripade, ze je potrebne zobrazit hviezdy
  if (cosSTheta<=0.1f) Color->A=cosSTheta/0.1f;
  // Skydome nebude nikdy uplne priesvitny
  if (cosSTheta<0.04f) Color->A=0.4f;
 }
private:
 float suntheta,T,suntheta2,suntheta3,T2;

 float Ax,Bx,Cx,Dx,Ex;
 float Ay,By,Cy,Dy,Ey;
 float AY,BY,CY,DY,EY;

 float xz,yz,X,Yz;

 float cosSTheta;
 float val1,val2;

 float fX,fY,fZ;
 float r,g,b;
 float expo;
 float cospsi,ac;

 float x,y,Y;
 float cosTheta,cosGamma;
 float gm;

 // Perez funkcie
 float PerezFunction_x(float cosTheta,float cosGamma,float gamma) {
  val1=(1.0f+Ax*exp(Bx/cosTheta))*(1.0f+Cx*exp(Dx*gamma)+Ex*(cosGamma*cosGamma));
  val2=(1.0f+Ax*exp(Bx))*(1.0f+Cx*exp(Dx*suntheta)+Ex*(cosSTheta*cosSTheta));
  return(val1/val2);
 }
 float PerezFunction_y(float cosTheta,float cosGamma,float gamma) {
  val1=(1.0f+Ay*exp(By/cosTheta))*(1.0f+Cy*exp(Dy*gamma)+Ey*(cosGamma*cosGamma));
  val2=(1.0f+Ay*exp(By))*(1.0f+Cy*exp(Dy*suntheta)+Ey*(cosSTheta*cosSTheta));
  return(val1/val2);
 }
 float PerezFunction_Y(float cosTheta,float cosGamma,float gamma) {
  val1=(1.0f+AY*exp(BY/cosTheta))*(1.0f+CY*exp(DY*gamma)+EY*(cosGamma*cosGamma));
  val2=(1.0f+AY*exp(BY))*(1.0f+CY*exp(DY*suntheta)+EY*(cosSTheta*cosSTheta));
  return(val1/val2);
 }

 // priemet parametrov x,y,Y na RGBA komponenty
 void ToRGBA(float x,float y,float Y,col4* rgb) {
  fX=x/y*Y;
  fY=Y;
  fZ=((1.0f-x-y)/y)*Y;

  r= 3.240479f*fX-1.537150f*fY-0.498535f*fZ;
  g=-0.969256f*fX+1.875991f*fY+0.041556f*fZ;
  b= 0.055648f*fX-0.204043f*fY+1.057311f*fZ;

  expo=-(1.0f/15000.0f);
  r=1.0f-exp(expo*r); g=1.0f-exp(expo*g); b=1.0f-exp(expo*b);

  if (r<0.0f) r=0.0f; if (g<0.0f) g=0.0f; if (b<0.0f) b=0.0f;
  if (r>1.0f) r=1.0f; if (g>1.0f) g=1.0f; if (b>1.0f) b=1.0f;

  rgb->R=r; rgb->G=g; rgb->B=b; rgb->A=1.0f;
 }
 float Angle(float theta1,float phi1,float theta2,float phi2) {
  cospsi=sin(theta1)*sin(theta2)*cos(phi2-phi1)+cos(theta1)*cos(theta2);
  if (cospsi>1.0f) return(0.0f);
  if (cospsi<-1.0f) return(PI);
  ac=acos(cospsi);
  return(ac);
 }
};

// trieda premennych a funkcii (objektov) pre pracu s oblohou
class CSkydome {
public:
 CSkydome();          // konstruktor
 virtual ~CSkydome(); // destruktor

 vec3 MoonPos,SunPos;

 // inicializacia oblohy
 int init(float Lat,float Long,float Time,float Day);
 // zrusenie oblohy
 void destroy(void);
 // aktualizacia oblohy
 void update(float deltatime);
 // vykreslenie oblohy
 void draw(GLfloat AT);
private:
 // polomer Skydomu, splostenie Skydomu
 float SkyRadius,Dampening;
 // pocet vrcholov, pruhov, stien, indikatorov
 unsigned int NumVertices,NumSlices,NumSides,NumIndices;
 // zemepisna vyska, zemepisna sirka, hodina, den
 float Latitude,Longitude,TimeOfDay,JulianDay;
 // velkost slnka, mesiaca
 float SunSize,MoonSize;

 vec3 *vertices;        // pole vrcholov
 float *VB;             // pole vrcholov pre VBO
 col4 *colors;          // pole farieb
 float *CB;             // pole farieb pre VBO
 unsigned int *indexes; // pole indexov
 unsigned int *IB;      // pole indexov pre VBO

 unsigned int VBObuffer;   // VBO buffer
 unsigned int VBOi_buffer; // VBO index buffer

 float polyAng,ang; // uhly
 float vx,vy,vz;    // vrcholy

 unsigned int i,j,k,ind; // pomocne premenne

 unsigned int SunTex,MoonTex,CloudsTex; // textura slnka, mesiaca, oblakov

 unsigned int NumStars; // pocet hviezd
 vec3 *stars;           // pole vrcholov - hviezd
 col4 *starscolor;      // pole jasu - hviezd
 float st,sp,c;         // uhly theta, phi a farba hviezdy

 // pre aktualizaciu oblohy
 float hour,solarTime,declination;
 float MoonTheta,MoonPhi;
 float SunTheta,SunPhi;
 float SliceAng,SideAng,theta,phi;

 float mat[16]; // pre ulozenie MODELVIEW matice

 vec3 vertx,verty; // pomocne vektory
 vec3 Vertices[4]; // pomocne vrcholy

 void RenderElements(bool moon); // vykreslenie mesiaca a hviezd / slnka
 void RenderSkydome(void);       // vykreslenie oblohy

 mSkycolor SkyShader; // priprava farby oblohy

 GLuint clouds_DL; // DisplayList pre objekt oblakov

 GLfloat clouds_time; // pre zmenu polohy oblakov
 GLuint CPsides;      // pocet poli objektu zakrivenej plochy
 GLfloat CPuv;        // pocet opakovani textury, ktorou je pokryta zakrivena plocha

 // velkost zakrivenej plochy, prirastok suradnic vrcholov, prirastok texturovacich suradnic
 GLfloat CPsize,Vdelta,Tdelta;
 // koeficient pre vypocet vysky vrcholu zakrivenej plochy, priesvitnosti
 GLfloat R1sq,Asq;

 GLuint a,b; // pomocne premenne

 clouds_plane CPv; // pomocna premenna
 clouds_plane *CPVertices; // pole vrcholov zakrivenej plochy
 GLuint CPVerticesNum;     // pocet vrcholov zakrivenej plochy
 GLuint *CPIndices;   // pole odkazov na konkretny vrchol
 GLuint CPIndicesNum; // pocet odkazov na konkretny vrchol

 // vytvorenie objektu oblakov
 int clouds_create(GLfloat R1,GLfloat R2,GLfloat A);
 // vykreslenie objektu oblakov
 void clouds_draw(GLfloat DT);
 // zrusenie objektu oblakov
 void clouds_destroy(void);
};

#endif
