// *************************************************************************
// * autor:   Marek Sopko                                                  *
// * kontakt: ... Banovce nad Bebravou, Slovenska republika                *
// *          http:   www.avatar.host.sk                                   *
// *          e-mail: avatar@host.sk                                       *
// *          mobil:  +421 907 952 453                                     *
// *                                                                       *
// * popis aplikacie: Engine v3.2                                          *
// *************************************************************************

// obsahom engine.cpp su nasledovne funkcie:
// - ukoncenie aplikacie
// - vytvorenie SDL okna
// - obsluha klavesnice
// - obsluha mysi
// - hlavna vykonavacia funkcia aplikacie

// hlavickove subory *******************************************************
#include "includes.h"
// vlastny .h subor
//#include "engine.h"

// hlavny interface aplikacie
#include "interface.h"

// *************************************************************************
// * deklaracie a definicie premennych a konstant                          *
// *************************************************************************
// sirka, vyska a farebna hlbka (pocet bitov na 1 pixel) kresliacej plochy
// pozn.: v X11, SDL_SetVideoMode nemeni rozlisenie obrazovky,
// vo Win32, ChangeDisplaySettings moze zmenit bpp
// scr_bpp=0 znamena, ze bude pouzita aktualna farebna hlbka obrazovky
int scr_w=800,scr_w2=scr_w/2;
int scr_h=600,scr_h2=scr_h/2;
//int scr_w=512,scr_w2=scr_w/2;
//int scr_h=384,scr_h2=scr_h/2;
int scr_bpp=0;

// znacky pre SDL_SetVideoMode
Uint32 videoFlags=0;
// celoobrazovkove zobrazenie
bool fullscr=false;
// informacie o SDL video subsysteme
const SDL_VideoInfo *videoInfo=NULL;
// vertikalna synchronizacia
int vsync=-1;
// celoobrazovkove vyhladzovanie
int fsaa=-1;
// SDL kresliaca plocha
SDL_Surface *surface=NULL;
// nazov okna a ikony
const char APP_title[64]="Engine v3.2 (OpenGL API)";
// opakovanie stlacenia klaves
bool keyrepeat=false;

// hlavny interface aplikacie - implementuje funkcie OpenGL kniznice
CInterface *interface=NULL;

// ukoncenie hlavneho cyklu
int isDone=0;
// obsluha udalosti
SDL_Event event;
// mys je/nie je aktivna
int grabmouse=0;
// pocas startu aplikacie sa nebude brat do uvahy poloha mysi
int ignoremouse=1;
// poloha mysky
int mox,moy;
// SDL okno je/nie je aktivne
int isActive=1;
// SDL last event type, last event button
int lasttype=0,lastbutton=0;
// SDL stav klaves
Uint8 *keystate=NULL;

// audio
Mix_Chunk *sounds[WAV_NUM];

// identifikator aktivneho a pasivneho avatara
// v tomto konkretnom pripade ide o vyber hraca - vojak/netvor
int ava=0; // 0,1 - aktivny je vojak, pasivny netvor
int avp=1; // 1,0 - aktivny je netvor, pasivny vojak

// *************************************************************************
// * definicie funkcii aplikacie                                           *
// *************************************************************************
// uvolnenie alokovanych zdrojov a obnovenie povodnych nastaveni obrazovky *
void APP_quit(int RETURNCODE)
{
 // zrusenie instancie triedy CInterface
 if (interface!=NULL) delete interface;
 interface=NULL;
 // zobrazenie kurzoru mysi
 SDL_ShowCursor(SDL_ENABLE);
 // audio
 loopi(WAV_NUM) { Mix_FreeChunk(sounds[i]); sounds[i]=NULL; }
 Mix_CloseAudio();
 // zrusenie SDL okna
 SDL_Quit();
 // ukoncenie aplikacie
 exit(RETURNCODE);
}

// inicializacia SDL *******************************************************
int SDL_initialize(void)
{
 // inicializacia SDL
 if (SDL_Init(SDL_INIT_TIMER|SDL_INIT_VIDEO|SDL_INIT_AUDIO)<0) {
  printf("Chyba inicializacie SDL: %s\n",SDL_GetError());
  APP_quit(1);
 }

 // znacky pre SDL_SetVideoMode
 videoFlags=SDL_OPENGL;                   // zapnutie OpenGL v SDL
 #if !defined(WIN32)
  videoFlags|=SDL_RESIZABLE;              // zmena velkosti SDL okna
 #endif
 if (fullscr) videoFlags|=SDL_FULLSCREEN; // celoobrazovkove zobrazenie
 //videoFlags|=SDL_GL_DOUBLEBUFFER;         // dvojity zasobnik pre prekreslovanie
 //videoFlags|=SDL_HWPALETTE;               // ulozenie palety do HW
 // ziskanie informacii o SDL video subsysteme
 videoInfo=SDL_GetVideoInfo();
 if (!videoInfo) {
  printf("Chyba ziskania informacii o SDL video subsysteme: %s\n",SDL_GetError());
  APP_quit(1);
 }
 // kontrola, ci mozno SDL kresliacu plochu umiestnit do HW
 //if (videoInfo->hw_available) videoFlags|=SDL_HWSURFACE;
 //else videoFlags|=SDL_SWSURFACE;
 // kontrola, ci mozno pracovat s HW blit
 // blitting - kopia casti zdrojovej textury do cielovej textury
 //if (videoInfo->blit_hw) videoFlags|=SDL_HWACCEL;

 // nastavenie atributov SDL kresliacej plochy
 // 8b pre cervenu, zelenu a modru farbu
 // 8b pre alfa kanal
 // 1 - dvojity zasobnik pre prekreslovanie
 // 16b zasobnik hlbky
 // 1b stencil zasobnik
 //SDL_GL_SetAttribute(SDL_GL_RED_SIZE,    8);
 //SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE,  8);
 //SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE,   8);
 //SDL_GL_SetAttribute(SDL_GL_ALPHA_SIZE,  8);
 SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER,1);
 // pozn.: SDL_DOUBLEBUFFER sa pouziva v SDL_GL_SetAttribute,
 // v SDL_SetVideoMode ovplyvni iba 2D prekreslovanie
 SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE,  16);
 SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE,1);
 // vertikalna synchronizacia
 #if SDL_VERSION_ATLEAST(1,2,11)
  if (vsync>=0) SDL_GL_SetAttribute(SDL_GL_SWAP_CONTROL,vsync);
 #endif
 // celoobrazovkove vyhladzovanie
 if (fsaa>0) {
  SDL_GL_SetAttribute(SDL_GL_MULTISAMPLEBUFFERS,1);
  SDL_GL_SetAttribute(SDL_GL_MULTISAMPLESAMPLES,fsaa);
 }

 // nastavenie SDL kresliacej plochy
 surface=SDL_SetVideoMode(scr_w,scr_h,scr_bpp,videoFlags);
 if (!surface) {
  fprintf(stderr,"Chyba nastavenia SDL kresliacej plochy: %s\n",SDL_GetError());
  APP_quit(1);
 }

 // INFO: sirka a vyska SDL kresliacej plochy
 printf("sirka: %i, vyska: %i\n",surface->w,surface->h);

 // sirka a vyska kresliacej plochy
 scr_w=surface->w; scr_w2=scr_w/2;
 scr_h=surface->h; scr_h2=scr_h/2;

 // prevzatie vstupu
 #ifdef WIN32
  SDL_WM_GrabInput(SDL_GRAB_ON);
 #else
  SDL_WM_GrabInput(fullscr ? SDL_GRAB_ON : SDL_GRAB_OFF);
 #endif

 // nastavenie nazvu okna a ikony
 SDL_WM_SetCaption(APP_title,NULL);
 // schovanie kurzoru mysi
 SDL_ShowCursor(SDL_DISABLE);
 // opakovanie stlacenia klaves
 SDL_EnableKeyRepeat(keyrepeat ? SDL_DEFAULT_REPEAT_DELAY : 0,
                     SDL_DEFAULT_REPEAT_INTERVAL);

 // audio
 if (Mix_OpenAudio(22050,AUDIO_S16,2,512)<0) {
  printf("Chyba inicializacie SDL_mixer: %s\n",Mix_GetError()); APP_quit(1);
 }
 int numtimesopened,nummusicdecoders,frequency,channels;
 Uint16 format;
 numtimesopened=Mix_QuerySpec(&frequency,&format,&channels);
 if (!numtimesopened) printf("Chyba nacitania informacii o audiu\n");
 else {
  char format_str[16];
  strcpy(format_str,"neznamy");
  switch (format) {
   case AUDIO_U8: strcpy(format_str,"U8"); break; 
   case AUDIO_S8: strcpy(format_str,"S8"); break;
   case AUDIO_U16LSB: strcpy(format_str,"U16LSB"); break;
   case AUDIO_S16LSB: strcpy(format_str,"S16LSB"); break;
   case AUDIO_U16MSB: strcpy(format_str,"U16MSB"); break;
   case AUDIO_S16MSB: strcpy(format_str,"S16MSB"); break;
  }
  nummusicdecoders=Mix_GetNumMusicDecoders();
  printf("audio otvorene %d krat, pocet dekoderov: %d\n",
         numtimesopened,nummusicdecoders);
  loopi(nummusicdecoders) printf("dekoder %d je urceny pre %s\n",i,Mix_GetMusicDecoder(i));
  printf("frekvencia: %dHz, format: %s, pocet kanalov: %d\n",
         frequency,format_str,channels);
 }
 sounds[MUSIC_WAV]=Mix_LoadWAV("data/sounds/wind.wav");
 if (!sounds[MUSIC_WAV]) { printf("Nemozno nahrat zvuk MUSIC\n"); APP_quit(1); }
 sounds[SHOT_WAV]=Mix_LoadWAV("data/sounds/gun.wav");
 if (!sounds[SHOT_WAV]) { printf("Nemozno nahrat zvuk SHOT\n"); APP_quit(1); }
 // hlasitost
 Mix_VolumeChunk(sounds[MUSIC_WAV],MIX_MAX_VOLUME/2);
 Mix_VolumeChunk(sounds[SHOT_WAV],MIX_MAX_VOLUME/2);

 return(0);
}

// zmena velkosti kresliacej plochy po zmene velkosti okna *****************
int WND_resize(int W,int H)
{
 // zmena velkosti SDL kresliacej plochy
 SDL_Surface *tmp_surface=SDL_SetVideoMode(W,H,scr_bpp,surface->flags);
 if (!tmp_surface) {
  fprintf(stderr,"Chyba nastavenia SDL kresliacej plochy po zmene velkosti okna: %s\n",SDL_GetError());
  APP_quit(1);
 }
 surface=tmp_surface;

 // sirka a vyska kresliacej plochy
 scr_w=surface->w; scr_w2=scr_w/2;
 scr_h=surface->h; scr_h2=scr_h/2;

 // inicializacia parametrov Avatara po zmene velkosti okna
 if (interface!=NULL) interface->avatar[ava].init();

 return(0);
}

// obsluha klavesnice ******************************************************
void APP_keyboard(void)
{
 keystate=SDL_GetKeyState(NULL); // nacitanie stavu klaves

 isDone=(int) keystate[SDLK_ESCAPE];                          // ESC
 // reset parametrov Avatara
 if (keystate[SDLK_i]) interface->avatar[ava].reset();
 if (keystate[SDLK_o]) interface->player[avp].setAnim(LEGS_IDLE,TORSO_STAND);
 // pohyb Avatara
 interface->avatar[ava].goForward=(GLfloat) keystate[SDLK_w];      // dopredu
 interface->avatar[ava].goBack=   (GLfloat) keystate[SDLK_s];      // dozadu
 interface->avatar[ava].goLeft=   (GLfloat) keystate[SDLK_a];      // dolava
 interface->avatar[ava].goRight=  (GLfloat) keystate[SDLK_d];      // doprava
 interface->avatar[ava].goUp=     (GLfloat) keystate[SDLK_r];      // hore
 interface->avatar[ava].goDown=   (GLfloat) keystate[SDLK_f];      // dole
 interface->avatar[ava].goLeanLeft= (GLfloat) keystate[SDLK_q];    // uklon dolava
 interface->avatar[ava].goLeanRight=(GLfloat) keystate[SDLK_e];    // uklon doprava
 interface->avatar[ava].goCrouch= (GLfloat) keystate[SDLK_c];      // vstat/cupnut
 interface->avatar[ava].goJump=   (GLfloat) keystate[SDLK_SPACE];  // skok
 interface->avatar[ava].goFast=   (GLfloat) keystate[SDLK_LSHIFT]; // chodza alebo beh
 // nastavenie prirastkov pre zmenu natocenia Avatara
 interface->avatar[ava].setrotzl((GLfloat) keystate[SDLK_q]);
 interface->avatar[ava].setrotzr((GLfloat) keystate[SDLK_e]);

 // zap/vyp casovo zavisleho pohybu objektov
 interface->pause_onoff=(bool) keystate[SDLK_p];
 // zap/vyp vypisu helpu
 interface->help_onoff=(bool) keystate[SDLK_h];
 // zap/vyp svietidla
 if (keystate[SDLK_x]) interface->change_flashlight_onoff();

 // vyplnene trojuholniky/iba hrany
 if (keystate[SDLK_l]) interface->change_polygon_mode();

 // menu
 if (keystate[SDLK_F1]) show_menu=!show_menu;
 if (show_menu) {
  // zobrazenie kurzoru mysi
  SDL_ShowCursor(SDL_ENABLE);
 }
 else {
  // schovanie kurzoru mysi
  SDL_ShowCursor(SDL_DISABLE);
 }

 // posun Avatara na zaklade vstupu z klavesnice
 interface->avatar[ava].setmovxyz();
}

// obsluha mysi ************************************************************
void APP_mouse(int DX,int DY)
{
 // nastavenie prirastkov pre zmenu natocenia Avatara
 interface->avatar[ava].setrotxy((GLfloat) DY,(GLfloat) DX);
}

// hlavna vykonavacia funkcia aplikacie ************************************
int main(int argc,char **argv)
{
 int x; // pomocna premenna

 printf("----------------------------------------------------------\n");

 // inicializacia SDL
 printf("SDL init\n");
 SDL_initialize();

 // vytvorenie instancie triedy CInterface
 interface=new CInterface();
 if (interface==NULL) {
  printf("Chyba tvorby instancie triedy CInterface!\n");
  APP_quit(1);
 }

 // inicializacia OpenGL
 printf("OpenGL init\n");
 if (interface->OGL_initialize()) APP_quit(1);

 // nastartovanie casovaca
 interface->resettime();

 // audio
 if (!Mix_Playing(MUSIC_WAV)) {
  Mix_PlayChannel(MUSIC_WAV,sounds[MUSIC_WAV],-1);
 }

 // hlavny programovy cyklus
 while (!isDone) {
  // ukladanie a spracovavanie udalosti
  lasttype=0; lastbutton=0;
  while (SDL_PollEvent(&event)) {
   switch (event.type) {
    case SDL_ACTIVEEVENT:     // kreslenie iba ak je okno aktivne
     if (event.active.state&SDL_APPINPUTFOCUS) grabmouse=event.active.gain;
     if (event.active.gain) { // ak dojde k aktivacii okna
      interface->resettime(); // restartuje sa casovac
      grabmouse=1;
      isActive=1;
      // prehravanie audia
      Mix_Resume(-1);
     } else {                 // okno nie je aktivne
        grabmouse=0;
        isActive=0;
        // pauza prehravania audia
        Mix_Pause(-1);
       }
    break;
    #if !defined(WIN32)
     case SDL_VIDEORESIZE: // zmena velkosti okna
      WND_resize(event.resize.w,event.resize.h);
     break;
    #endif
    case SDL_QUIT:        // zrusenie SDL okna
     isDone=1;
    break;
    case SDL_KEYDOWN:     // obsluha klavesnice
    case SDL_KEYUP:
     APP_keyboard();
    break;
    case SDL_MOUSEMOTION: // obsluha mysi
     if (ignoremouse) { ignoremouse--; break; }
     if (show_menu) { mox=event.motion.x; moy=event.motion.y; break; }
     #ifndef WIN32
      if (!(surface->flags&SDL_FULLSCREEN) && grabmouse) {
       if (event.motion.x==scr_w2 && event.motion.y==scr_h2) break;
       SDL_WarpMouse(scr_w2,scr_h2);
      }
      if ((surface->flags&SDL_FULLSCREEN) || grabmouse)
     #endif
       APP_mouse(event.motion.xrel,event.motion.yrel);
    break;
    case SDL_MOUSEBUTTONDOWN:
    case SDL_MOUSEBUTTONUP:
     if (!show_menu) {
      if (lasttype==event.type && lastbutton==event.button.button) break;
      // stlacenie laveho a praveho tlacidla mysi
      interface->avatar[ava].isFire1=(GLubyte) SDL_GetMouseState(NULL,NULL)&SDL_BUTTON(SDL_BUTTON_LEFT);
      interface->avatar[ava].isFire2=(GLubyte) SDL_GetMouseState(NULL,NULL)&SDL_BUTTON(SDL_BUTTON_RIGHT);
      lasttype=event.type; lastbutton=event.button.button;
     }
     else {
	  if ((GLubyte) SDL_GetMouseState(NULL,NULL)&SDL_BUTTON(SDL_BUTTON_LEFT)) {
       interface->menu.update(mox,moy);
      }
	 }
    break;
    default: break;
   }
  }
  // aktualizacia a vykreslenie sceny
  if (isActive) {
   // aktualizacia parametrov Avatara
   interface->avatar[ava].update();
   x=otree.col_tris_num;
   interface->avatar[avp].update();
   otree.col_tris_num=x;
   // casovo zavisla uprava objektov, resp. sceny
   interface->OGL_idle();
   // vykreslenie OpenGL sceny
   interface->OGL_draw();
   // vymena zobrazovacich zasobnikov => zobrazenie sceny
   SDL_GL_SwapBuffers();
  }
  // aktualizacia casovaca
  interface->updatetime();
 }

 // ukoncenie aplikacie
 APP_quit(0);
}
